#include "min-s-vps.hpp"
#include <iostream>
using namespace std;

/** The parameter \p is the number of steps used when sampling over a complete
    parameter range , e.g. the parameter a_0 is sampled over 0, 1/(n_steps-1),
    2/(n_steps-1), ..., 1. The parameter \p N is the number of values in the
    vector \t V containing the approximation for the superoptimal parameters.
    Each entry in \t V can either be empty or contain a set of values a_0, a_1,
    b_0, b_1, beta_0, beta_1, gamma_0, gamma_1, s, and vp. The entries are
    sorted according to the deviation range they fall into (n=N-1):\n
    V_0: vp \in [-0.5/n, 0.5/n), \n
    V_1: vp \in [0.5/n, 1.5/n), \n
    V_2: vp \in [1.5/n, 2.5/n), \n
    ..., \n
    V_n: vp \in [(n-0.5)/n, (n+0.5)/n).\n
    Initially, all entries are empty. */
min_s_vps_tp::min_s_vps_tp(int n_steps, int N) :
    n_steps_mb(n_steps), V(N, (s_vp_tp *)0)
{}

min_s_vps_tp::~min_s_vps_tp() {
    int maxstep = n_steps_mb-1;
    for (int i = 0; i <= maxstep; ++i)
        delete V[i];
}

/** First, this function calculates the disturbance/deviation pair from the
    parmameters \p i_a_0, \p i_b_0, \p i_b_0, \p i_beta_1, \p i_gamma_0, and \p
    i_gamma_1, and \p maxstep. Then, it inserts the resulting set into the
    corresponding entry in the vector \t V either

    - if the entry is empty, or
    - if the calculated disturbance is smaller than the disturbnace of the
      already existing entry. */
void min_s_vps_tp::
try_to_store(int i_a_0, int i_b_0, int i_beta_0, int i_beta_1,
             int i_gamma_0, int i_gamma_1, int maxstep)
{
    s_vp_tp *s_vp = new s_vp_tp(i_a_0, i_b_0, i_beta_0, i_beta_1,
                                i_gamma_0, i_gamma_1, maxstep);
    int slot = round_to_int(s_vp->vp()*(V.size()-1));
    if (V[slot] == 0) {
        V[slot] = s_vp;
    } else if (s_vp->s() < V[slot]->s()) {
        delete V[slot];
        V[slot] = s_vp;
    } else
        delete s_vp;
}

/** Samples over all valid parameter combinations (...) \in {...}^6.  Each
    combination is tried to be inserted into the vector \t V. This vector
    eventually contains a list of parameter combinations accounting for the
    lowest disturbance in each deviation range. That list, obviously, is an
    approximation for the superoptimal parameters of the discussed estimation
    procedure.
 
    Simple status information about the progress of the run is output to
    stderr. */
void min_s_vps_tp::sample() {
    int maxstep = n_steps_mb-1;

    cerr << n_steps_mb << " steps:" << flush;
    for (int i_a_0 = 0; i_a_0 <= maxstep; ++i_a_0) {
        cerr << " " << i_a_0 << flush;
        for (int i_b_0 = 0; i_b_0 <= maxstep; ++i_b_0) {
            for (int i_beta_0 = 0; i_beta_0 <= maxstep; ++i_beta_0) {
                for (int i_beta_1 = 0; i_beta_1 <= maxstep; ++i_beta_1) {
                    for (int i_gamma_0 = 0; i_gamma_0 <= maxstep; 
                         ++i_gamma_0) 
                    {
                        if (i_a_0 != i_b_0) {
                            try_to_store(i_a_0, i_b_0, i_beta_0, i_beta_1,
                                         i_gamma_0, i_gamma_0, maxstep);
                            try_to_store(i_a_0, i_b_0, i_beta_0, i_beta_1,
                                         i_gamma_0, maxstep-i_gamma_0, 
                                         maxstep);
                        } else {
                            for (int i_gamma_1 = 0; i_gamma_1 <= maxstep; 
                                 ++i_gamma_1)
                            {
                                try_to_store(i_a_0, i_b_0, i_beta_0, i_beta_1,
                                             i_gamma_0, i_gamma_1, maxstep);
                            }
                        }
                    }
                }
            }
        }
    }                            
    cerr << endl;
}

/** Outputs parameters of interest in the non empty entries the vector \t V to
    stdout. */
void min_s_vps_tp::output() {
    cout << "# entry vp s a_0 a_1 b_0 b_1 beta_0 beta_1 gamma_0 gamma_1" 
         << endl;
    for (int i = 0; i <= V.size(); ++i)
        if (V[i] != 0)
            cout << i << "\t"
                 << V[i]->vp() << "\t"
                 << V[i]->s() << "\t"
                 << V[i]->a_0() << "\t"
                 << V[i]->a_1() << "\t"
                 << V[i]->b_0() << "\t"
                 << V[i]->b_1() << "\t"
                 << V[i]->beta_0() << "\t"
                 << V[i]->beta_1() << "\t"
                 << V[i]->gamma_0() << "\t"
                 << V[i]->gamma_1()
                 << endl;
}
