#include "s-vp.hpp"
#include <cmath>
using namespace std;

/** Creates the set (see the documentation of this class for more information)
    from the parameters a_0 = \p i_a_0 / \p maxstep, etc.. */
s_vp_tp::s_vp_tp(int i_a_0, int i_b_0, int i_beta_0, int i_beta_1,
                 int i_gamma_0, int i_gamma_1, int maxstep) :
    maxstep_mb(maxstep)
{
    // The integer values are stored so that they can later be used e.g. for a
    // precise comparison of parameters:
    i_a_mb[0] = i_a_0;
    i_b_mb[0] = i_b_0;
    i_beta_mb[0] = i_beta_0;
    i_beta_mb[1] = i_beta_1;
    i_gamma_mb[0] = i_gamma_0; 
    i_gamma_mb[1] = i_gamma_1;

    a_mb[0] = double(i_a_0) / maxstep;
    b_mb[0] = double(i_b_0) / maxstep;
    beta_mb[0] = double(i_beta_0) / maxstep;
    beta_mb[1] = double(i_beta_1) / maxstep;
    gamma_mb[0] = double(i_gamma_0) / maxstep;
    gamma_mb[1] = double(i_gamma_1) / maxstep;
    
    // The parameters a_1 and b_1 (see 0309302232):
    if (i_a_0 != i_b_0) {
        if (i_gamma_1 == i_gamma_0) {
            a_mb[1] = sqrt(1-pow(a_mb[0],2));
            b_mb[1] = sqrt(1-pow(b_mb[0],2));
        } else {
            a_mb[1] = sqrt(1-pow(b_mb[0],2));
            b_mb[1] = sqrt(1-pow(a_mb[0],2));
        }
    } else {
        a_mb[1] = b_mb[1] = sqrt(1-pow(a_mb[0],2));
    }

    // The disturbance and deviation
    for (int m = 0; m <= 1; ++m)
        update_P(m);
    update_s();
    update_vp();
}

/** Returns true, if a comparison "a_m = b_m" in the notation of 0309302232
    would return true. */
bool s_vp_tp::a_m_eq_b_m(int m) {
    return i_a_mb[0] == i_b_mb[0];
}

/** Returns true, if a comparison "a_m = 0" in the notation of 0309302232 would
    return true. */
bool s_vp_tp::a_m_eq_0(int m) {
    return
        (m == 0 && i_a_mb[0] == 0) || 
        (m == 1 &&           
         ((i_a_mb[0] == maxstep_mb && i_gamma_mb[1] == i_gamma_mb[0]) ||
          (i_b_mb[0] == maxstep_mb && i_gamma_mb[1] != i_gamma_mb[0])));
}

/** Returns true, if a comparison "b_m = 0" in the notation of 0309302232 would
    return true. */
bool s_vp_tp::b_m_eq_0(int m) {
    return 
        (m == 0 && i_b_mb[0] == 0) ||
        (m == 1 &&          
         ((i_b_mb[0] == maxstep_mb && i_gamma_mb[1] == i_gamma_mb[0]) ||
          (i_a_mb[0] == maxstep_mb && i_gamma_mb[1] != i_gamma_mb[0])));
}

/** Returns true, if the comparison "a_m = b_m = 0" in the notation of
    0309302232 would return true. */
bool s_vp_tp::a_m_eq_b_m_eq_0(int m) {
    return i_a_mb[0] == i_b_mb[0] && a_m_eq_0(m);
}

/** Updates \p P_mb[m] (see 0309302232). */
void s_vp_tp::update_P(int m) {
    double a = a_mb[m], b = b_mb[m], beta = beta_mb[m];
    double &P = P_mb[m];
    if (a_m_eq_b_m(m))
        P = 1./3. * pow(a,2);
    else if (a_m_eq_0(m))
        P = 0.5 * pow(b,2) * pow(beta-1,2);
    else if (b_m_eq_0(m))
        P = 0.5 * pow(a,2) * pow(beta,2);
    else
        P = 1. / (2.*pow(pow(a,2)-pow(b,2),2)) * 
            (pow(a,6) * pow(beta,2) + 
             pow(b,6) * pow(1-beta,2) +
             pow(a,4) * pow(b,2) * beta * (4-7*beta) +
             pow(a,2) * pow(b,4) * (1-beta) * (7*beta-3)) +
            2.*pow(a,4)*pow(b,4) / pow(pow(a,2)-pow(b,2),3) *
            (1 - 6*beta*(1-beta)) * log(a/b);
}

/** Updates the disturbance \p s_mb (see 0309302232). */
void s_vp_tp::update_s() {
    s_mb = 0.;
    for (int m = 0; m <= 1; ++m) {
        double a = a_mb[m], b = b_mb[m], beta = beta_mb[m], 
            gamma = gamma_mb[m];
        s_mb += P_mb[m];
        s_mb -= 1./3. * 
            (pow(a,2)*beta*(1+gamma) +
             pow(b,2)*(1-beta)*(2-gamma) +
             2*a*b*sqrt(beta*(1-beta))*sqrt(gamma*(1-gamma)));
    }
    s_mb += 1./3.;
}

/** Updates the postestimation deviation \p vp_mb (see 0309302232). */
void s_vp_tp::update_vp() {
    vp_mb = 0.;
    for (int m = 0; m <= 1; ++m) {
        double a = a_mb[m], b = b_mb[m], beta = beta_mb[m];
        vp_mb += P_mb[m];
        if (!a_m_eq_b_m_eq_0(m))
            vp_mb -= 0.5 *
                pow(pow(a,2)*beta + pow(b,2)*(1-beta), 2) / 
                (pow(a,2)+pow(b,2));
    }
}
