#include <power.hpp>
#include <sstream>
#include <cmath>
#include <iostream>
#include <cstdlib>
#include <limits>
using namespace std;

double pi = 2*acos(0.);

/// Default values for \p eps for relative comparison in approx_eq.
const double default_eps_r = 1e-5;

/// Default values for \p eps for absolute comparison in approx_eq.
const double default_eps_a = 1e-5;

enum approx_tp { aprx_r, aprx_a };

/** Compares the floating point values and \p x and \p y and returns true only
    if they are closer to each other than

    - eps*min(x, y) (\p comp = aprx_r = relative comparison), or
    - eps           (\p comp = aprx_a = absolute comparison).

    If \p eps < 0 then \p default_eps_r respectively \p default_eps_a is used
    instead. */
inline bool 
approx_eq(double x, double y, approx_tp comp = aprx_r, 
          double eps = -1)
{
    using namespace std;
    if (comp == aprx_a) {
        if (eps < 0) 
            eps = default_eps_a;
        return abs(x - y) <= eps;
    } else {
        if (eps < 0) 
            eps = default_eps_r;
        return abs(x - y) <= eps * min(abs(x), abs(y));
    }
}

typedef double (*sample_function_tp)(double);

/** Fills the record stream \p rcrd_strm with \p N samples f(0*Delta_t*2*pi),
    f(1*Delta_t*2*pi), ..., f((N-1)*Delta_t*2*pi).  */
void sample(stringstream &rcrd_strm, sample_function_tp f, unsigned long N, 
            double Delta_t) 
{
    rcrd_strm.clear();
    rcrd_strm.str("");
    rcrd_strm << "# t |c_1_tilde|^2 |c_1|^2 g_1\n" << endl;
    for (int i = 0; i < N; ++i)
        rcrd_strm << "0 1 " << f(i*Delta_t) << " 3\n";
    rcrd_strm << flush;
}

/** Checks by assertion that the power spectrum contained in \p power is
    approximately (i.e. without taking into account numerical inaccuracies) to
    that given by \p power_expected and \p power_expected_size. The stream \p
    power may be completely emptied during that procedure. */
void check_power(stringstream &power_strm, double power_expected[][2], 
                 unsigned long power_expected_size) 
{
    typedef double double2[2];
    double2 *power = new double2[power_expected_size];
    double tmp;
    unsigned long n = 0;
    power_strm.ignore(numeric_limits<int>::max(), '\n'); // Ignores first line
    while ((power_strm >> tmp) && (n/2 < power_expected_size))
        power[n/2][n%2] = tmp, ++n;
    assert(n == 2*power_expected_size);

    for (unsigned long i = 0; i < power_expected_size; ++i) {
        assert(approx_eq(power[i][0], power_expected[i][0], aprx_a, 1e-3));
        assert(approx_eq(power[i][1], power_expected[i][1], aprx_a, 1e-3));
    }
    delete[] power;
}

/** Tests miscellaneous stuff. */
void test1() {
    // QQQ: test totally empty data, zero line data, one line data, and other
    // pathologic cases
}

double test2_0_5(double t) { return 0.5; }

double test2_shifted_Gaussian(double t) { return exp(-pow(3-t, 2)); }

/** Tests creation of the power spectrum. */
void test2() {
    stringstream rcrd_strm, power_strm;

    // sin sampled 4 times during each of 25 periods:
    sample(rcrd_strm, sin, 100, 0.25*2.*pi);
    rcrd_to_powr(100, 51, 0.25*2.*pi, rcrd_strm, power_strm);
#include "test2-1_data.cpp"
    check_power(power_strm, power2_1_expected, 51);

    // sin sampled 100 times during 1 period:
    sample(rcrd_strm, sin, 100, 0.01*2.*pi);
    power_strm.str(""), power_strm.clear();
    rcrd_to_powr(100, 51, 0.01*2.*pi, rcrd_strm, power_strm);
#include "test2-2_data.cpp"
    check_power(power_strm, power2_2_expected, 51);

    // The constant 0.5 sampled 19 times:
    sample(rcrd_strm, test2_0_5, 19, 1.);
    power_strm.str(""), power_strm.clear();
    rcrd_to_powr(19, 10, 1., rcrd_strm, power_strm);
#include "test2-3_data.cpp"
    check_power(power_strm, power2_3_expected, 10);

    // Shifted Gaussian sampled 12 times:
    sample(rcrd_strm, test2_shifted_Gaussian, 12, 0.413);
    power_strm.str(""), power_strm.clear();
    rcrd_to_powr(12, 7, 0.413, rcrd_strm, power_strm);
#include "test2-4_data.cpp"
    check_power(power_strm, power2_4_expected, 7);

    // Shifted Gaussian whose main part is only partially sampled (12 times):
    sample(rcrd_strm, test2_shifted_Gaussian, 12, 0.25);
    power_strm.str(""), power_strm.clear();
    rcrd_to_powr(12, 7, 0.25, rcrd_strm, power_strm);
#include "test2-5_data.cpp"
    check_power(power_strm, power2_5_expected, 7);
}

/** Tests condensing of the power spectrum. */
void test3() {
    stringstream rcrd_strm, power_strm;

    // power_N too large:
    sample(rcrd_strm, sin, 100, 1.);
    try {
        rcrd_to_powr(100, 52, 1., rcrd_strm, power_strm);
        assert(false);
    } catch (const out_N_too_large_error_tp &e) {
        assert(strcmp(e.what(), "The number of output values in the condensed first part of the power spectrum,\nout_N, needs to be smaller or equal to N/2+1, where N is the number of samples.") == 0);
    }        

    // power_N with maximum size:
    sample(rcrd_strm, sin, 100, 1.);
    rcrd_to_powr(100, 51, 1., rcrd_strm, power_strm);

    // power_N = 0:
    sample(rcrd_strm, sin, 100, 1.);
    power_strm.str("");
    rcrd_to_powr(100, 0, 1., rcrd_strm, power_strm);
    assert(power_strm.str() == "# omega/(dim(Delta_t)^-1) amplitude\n");

#if 0 // QQQ
    for (int i = 0; i < 100; ++i)
        rcrd_strm << "0 1 " << sin((i/100.)*(2.*pi)) << " 3\n";
    try {
        rcrd_to_powr(100, 0, 1., rcrd_strm, power_strm);
        assert(false);
    } catch (const power_N_too_large_error_tp &e) {
        assert(strcmp(e.what(), "The number of output values in the power "
                      "spectrum, power_N, needs to be smaller\nor equal to "
                      "N/2+1, where N is the number of samples.") == 0);
    }
#endif 
}

#if 0
void qqq() {
    cout << "-----------" << endl;
    double tmp;
    while (power_strm >> tmp)
        cout << tmp << endl;
    rcrd_strm << flush;
    cout << "rcrd_strm = " << rcrd_strm.str() << "\n#" << endl;
}
#endif

int main() {
    cout << "Testing ... " << endl;
    try {
        test1();
        test2();
        test3();
    } catch (const exception &e) {
        cout << "failed." << endl;
        cerr << e.what() << endl;
        return EXIT_FAILURE;
    }
    cout << "passed." << endl;
    return 0;
}
