#include "frequency-tracking.hpp"
#include <cmath>
using namespace std;

// QQQ: Check units everywhere

frequency_tracking_tp::frequency_tracking_tp(params_tp &params, 
                                             std::ostream &misc_stream, 
                                             std::ostream &rcrd_stream) :
    tracking_tp(params, misc_stream, rcrd_stream, "frequency-tracking"),
    measurement_mb(measurement_operators()),
    Delta_t_mb(params.get<double>(sec_name_mb+":Delta_t")),
    N_mb(params.get<int>(sec_name_mb+":N")),
    omega_tilde_aprx_mb(0), beta_g_mb(0.5),
    n_rec_points_mb(params.get<int>(sec_name_mb+":n_rec_points",1)),
    U_mb(1,0,0,1),
    U_half_mb(1,0,0,1),
    U_rec_mb(1,0,0,1),
    phase_g_mb(0)
{
    if (E_1_mb < E_0_mb) // QQQ: is this check necessary? Or even make this general check?
        throw E_1_less_than_E_0_error_tp();
    if (Delta_t_mb <= 0)
        throw Delta_t_too_small_error_tp();
    if (N_mb < 1)
        throw N_too_small_error_tp();
    // QQQ: check if Delta_t < pi/omega_max; This is damn important; Does this
    // check make sense? No ?

    U_mb = H_to_unitary(tracking_tp::E_0(), tracking_tp::E_1(), 
                        tracking_tp::E_state_0(), tracking_tp::E_state_1(),
                        Delta_t_mb);
    U_half_mb = H_to_unitary(tracking_tp::E_0(), tracking_tp::E_1(), 
                             tracking_tp::E_state_0(), 
                             tracking_tp::E_state_1(),
                             Delta_t_mb/2);

    if (n_rec_points_mb < 1)
        // n_rec_points_mb needs to be at least once because otherwise values of
        // |c_1|^2, guesses, etc. are lost
        throw n_rec_points_too_small_error_tp();
    Delta_t_rec_mb = (Delta_t_mb/2) / n_rec_points_mb;
    U_rec_mb = H_to_unitary(tracking_tp::E_0(), tracking_tp::E_1(), 
                            tracking_tp::E_state_0(), tracking_tp::E_state_1(),
                            Delta_t_rec_mb);

    try {
        dont_record_mb = params.get_vector<int>(sec_name_mb+":dont_record");
    } catch(...) {
        // By default, record all data:
        dont_record_mb.push_back(-1), dont_record_mb.push_back(-1);
    }
    if (dont_record_mb.size() == 1)
        // Only the first measurement was specified => omit everything till the
        // end:
        dont_record_mb.push_back(N_mb);

    g_1_mb = 1 - beta_g_mb + beta_g_mb*cos(phase_g_mb);

    stream_misc1();

    try {
        recording_mb.reserve(N_mb * 3*n_rec_points_mb);
    } catch(...) {
        throw not_enough_memory_error_tp();
    }
}

/** QQQ
    Evolves the system by a single recording point. */
void frequency_tracking_tp::evolve_rec() {
    // Evolve time:
    t_mb += Delta_t_rec_mb;

    // Evolve and renormalize (to anticipate numerical instability) the
    // disturbed and undisturbed state:
    psi_mb = U_rec_mb * psi_mb;
    psi_mb.normalize();
    psi_tilde_mb = U_rec_mb * psi_tilde_mb;
    psi_tilde_mb.normalize();
    
    // Evolve guess:
    phase_g_mb += omega_tilde_aprx_mb * Delta_t_rec_mb;
    g_1_mb = 1 - beta_g_mb + beta_g_mb*cos(phase_g_mb);
}

/** QQQ 

    Evolves the system by Delta_t if half=false and by Delta_t/2 if
    half=true. */
void frequency_tracking_tp::evolve(bool half) {
    // Evolve time:
    double tau = half ? Delta_t_mb/2 : Delta_t_mb;
    t_mb += tau;

    // Evolve and renormalize (to anticipate numerical instability) the
    // disturbed and undisturbed state:
    operator_tp &U = half ? U_half_mb : U_mb;
    psi_mb = U * psi_mb;
    psi_mb.normalize();
    psi_tilde_mb = U * psi_tilde_mb;
    psi_tilde_mb.normalize();
    
    // Evolve guess:
    phase_g_mb += omega_tilde_aprx_mb * tau;
    g_1_mb = 1 - beta_g_mb + beta_g_mb*cos(phase_g_mb);
}

/** QQQ 
    last_measurement = "number of the last measurement" (-1 means no measurement at t=0)
    (QQQ: actually -1 is the measurement before tracking started - see QQQcalc-#)
*/
void frequency_tracking_tp::record_and_evolve(bool half, int last_measurement) {
    if (last_measurement >= 2*dont_record_mb[0]-1 && 
        last_measurement < 2*dont_record_mb[1]-1) 
    {
        evolve(half);
    } else {
        for (int i = 0; i < (half ? n_rec_points_mb : 2*n_rec_points_mb); ++i) {
            record(); // We record before evolving the system so that the data
                      // at t=0 can be recorded
            evolve_rec();
        }
    }
}

/** QQQ 

    Returns the number of measurements performed. */
unsigned long frequency_tracking_tp::track() {
    //QQQ: move otu of tracking: delay();
    vector<operator_tp> Ms;
    Ms.push_back(operator_tp(1, 0, 0, 0));
    Ms.push_back(operator_tp(0, 0, 0, 1));
    measurement_tp measurement(Ms);

    int n_measurements = 0;
    M_mb[0] = M_mb[1] = 0;
    
    // QQQ: record at t = 0?

    // Note that we skip the initial measurement explained in QQQ because we
    // don't need it in the simulation (we can just set psi_mb = one_mb - see
    // below)
    ++ n_measurements;

    // Simulates the tracking procedure and records the corresponding data at
    // each recording point:
    for (int N = 1; N <= N_mb; ++N) { // QQQ: replace N_mb by N_max_mb and update
                                      // corresponding input files?
        for (int half = 0; half <= 1; ++half) {
            // Reset |psi> to |1> (in reality this can be done by applying a uniform
            // transformation to psi that depends on the last measurement result:
            psi_mb = one_mb;

            // Wait for Delta_t if half=false, or for Delta_t/2 if half = true,
            // and update and record data during that time:
            record_and_evolve(half, n_measurements-2);
            
            // Perform measurement and update approximation for omega_tilde:
            int outcome = measurement_mb.measure(psi_mb);
            ++ n_measurements;
            if (outcome == 1)
                ++ M_mb[half?0:1];
        }
        update_omega_tilde_aprx(N);
    }
    stream_misc2(); // QQQ: this will be appended after each tracking run!
    return n_measurements;
}

/** See QQQ */
double frequency_tracking_tp::acos_f(double a, double b) {
    if (a >= 0 && a < 1 && b >= 4*a-3 && b <= 1) {
        double tmp = (b-1.)/(2.*(a-1.))-1.;
        if (tmp >= 1.) 
            // This case might be reached if tmp is slightly larger than 1.. In
            // this case, acos(tmp) might return NaN. So we treat this case
            // specially. Note that on some systems (e.g. on Intel systems),
            // acos(tmp) sometimes returns NaN, although the expression (tmp ==
            // 1.) returns true.
            return 0;
        else if (tmp <= -1.)
            // For an explanation, see above
            return pi;
        else
            return acos(tmp);
    } else
        return 0;
}

/** QQQ */
void frequency_tracking_tp::update_omega_tilde_aprx(int N) {
    double a_aprx = double(M_mb[0]) / N;
    double b_aprx = double(M_mb[1]) / N;
    omega_tilde_aprx_mb = 2/Delta_t_mb * acos_f(a_aprx, b_aprx);
}

/** Returns the amplitude of the undisturbed oscillation. */
double frequency_tracking_tp::beta() {
    return 
        2*abs((one_mb*E_state_0_mb)*(E_state_0_mb*psi_init_mb)*(psi_init_mb*one_mb)
              - norm((one_mb*E_state_0_mb)*(E_state_0_mb*psi_init_mb)));
}

/** Returns the frequency of the oscillation as defined by E_0_mb and
    E_1_mb. Therefore that frequency might be unequal to zero even if the
    amplitude of the oscillation is zero. */
double frequency_tracking_tp::omega() {
    return E_1_mb - E_0_mb; // Note that dim(E_j_mb) = h_bar/s (QQQ: really use absolute time?)
}

/** Returns the frequency of the oscillation (if the amplitude is zero then the
    frequency zero is returned). dim(omega_tilde) = Hz. Note that the disturbed
    and the undisturbed oscillation have the same frequency. */
double frequency_tracking_tp::omega_tilde() {
    if (approx_eq(beta(), 0, aprx_a))
        return 0;
    else
        return omega();
}

/** Approximation for omega_tilde() that was found in the first stage of
    tracking. If this function is called before the tracking procedure, the
    returned value is undefined. */
double frequency_tracking_tp::omega_tilde_aprx() {
    return omega_tilde_aprx_mb;
}

/** Returns a vector with the measurement operators |0><0| and |1><1|. This
    function is needed for the initialization of measurement_mb. */
const std::vector<operator_tp> frequency_tracking_tp::measurement_operators() {
    vector<operator_tp> ops;
    ops.push_back(operator_tp(1,0,0,0));
    ops.push_back(operator_tp(0,0,0,1));
    return ops;
}

/** Appends miscellaneous data to the stream \o misc_stream. In this run some 
    data that is available before tracking is appended. */
void frequency_tracking_tp::stream_misc1() {
    misc_stream_mb << "\n\n";
    misc_stream_mb << "[frequency-tracking]\n";
    misc_stream_mb << "\n";
    misc_stream_mb << "omega = " << omega()
                   << " # Oscillation frequency as defined by E_0 and E_1\n"; // QQQ: dim?
    misc_stream_mb << "beta = " << beta()
                   << " # Amplitude of the undisturbed osciallation\n"; // QQQ: dim?
    misc_stream_mb << "omega_tilde = " << omega_tilde()
                   << " # 0 if beta = 0, omega if beta > 0\n"; // QQQ: dim?
}

/** Appends miscellaneous data to the stream \o misc_stream. In this run some
    data that is available after tracking is appended. */
void frequency_tracking_tp::stream_misc2() {
    misc_stream_mb << "omega_tilde_aprx = " << omega_tilde_aprx()
                   << " # Approximated value (by measurements) of omega_tilde\n"; // QQQ: dim?
}
