#ifndef TRASIM__H2_HPP
#define TRASIM__H2_HPP

#include "misc.hpp"
#include <string>
#include <complex>
#include <vector>
#include <stdexcept>
#include <cmath>

class state_tp;

/** Operator acting on states in 2D Hilbert space. 

    \author Felix E. Klee (felix.klee@inka.de)
    \date 2004 */
class operator_tp {
public:
    inline operator_tp(const operator_tp &);
    inline operator_tp(const complex_double_vect &);
    inline operator_tp(complex_double, complex_double, 
                       complex_double, complex_double);
    static const operator_tp zero; ///< Zero operator
    static const operator_tp one; ///< Unit operator
    inline complex_double &operator()(int, int);
    inline complex_double operator()(int, int) const;
    inline complex_double expectation_value(const state_tp &) const;
    inline bool unitary(double) const;
    inline complex_double_vect eigenvalues() const;
    class init_vector_size_error_tp;
private:
    complex_double A_mb[4];
};

class operator_tp::init_vector_size_error_tp : public std::runtime_error {
public:
    init_vector_size_error_tp() : 
        runtime_error("Initialization vector has wrong size.") {}
};

/** State in 2D Hilbert space. 

    \author Felix E. Klee (felix.klee@inka.de)
    \date 2004 */
class state_tp {
public:
    inline state_tp(complex_double, complex_double);
    inline state_tp(const complex_double_vect &);
    inline complex_double &operator()(int);
    inline complex_double operator()(int) const;
    inline bool normalized(double) const;
    inline void normalize();
    class init_vector_size_error_tp;
private:
    complex_double psi_mb[2];
};

class state_tp::init_vector_size_error_tp : public std::runtime_error {
public:
    init_vector_size_error_tp() : 
        runtime_error("Initialization vector has wrong size.") {}
};

operator_tp adj(const operator_tp &);
operator_tp operator*(const operator_tp &, const operator_tp &);
bool approx_eq(const operator_tp &, const operator_tp &,
               approx_tp, double eps);
inline operator_tp outer_product(const state_tp &, const state_tp &);

operator_tp::
operator_tp(const operator_tp &src) {
    A_mb[0] = src.A_mb[0], A_mb[1] = src.A_mb[1];
    A_mb[2] = src.A_mb[2], A_mb[3] = src.A_mb[3];
}

operator_tp::
operator_tp(const complex_double_vect &A_vect) {
    if (A_vect.size() != 4)
        throw init_vector_size_error_tp();
    A_mb[0] = A_vect[0], A_mb[1] = A_vect[1];
    A_mb[2] = A_vect[2], A_mb[3] = A_vect[3];
}

operator_tp::
operator_tp(complex_double A_00, complex_double A_01, 
            complex_double A_10, complex_double A_11)
{
    A_mb[0] = A_00, A_mb[1] = A_01, A_mb[2] = A_10, A_mb[3] = A_11;
}

/** Returns <i|A|j>, where A is this operator. */
complex_double &operator_tp::
operator()(int i, int j) {
    return A_mb[2*i+j];
}

/** Returns <i|A|j>, where A is this operator. */
complex_double operator_tp::
operator()(int i, int j) const {
    return A_mb[2*i+j];
}

/** Returns the expectation value of this operator for state \p psi. */
complex_double operator_tp::
expectation_value(const state_tp &psi) const {
    const operator_tp &A = *this;
    return 
        conj(psi(0))*A(0,0)*psi(0) + conj(psi(0))*A(0,1)*psi(1) +
        conj(psi(1))*A(1,0)*psi(0) + conj(psi(1))*A(1,1)*psi(1);
}

/** Returns true iff the operator is unitary. \p eps specifies the tolerance for
    numerical comparisons (-1 means to use the default value). 

    \sa approx_eq(double, double, approx_tp, double) */
bool operator_tp::unitary(double eps = -1) const {
    const operator_tp &U = *this;
    return approx_eq(adj(U)*U, operator_tp::one, aprx_a, eps);
}

/** Returns the vector of eigenvalues of the operator. */
complex_double_vect operator_tp::eigenvalues() const {
    const operator_tp &A = *this;
    complex_double_vect lambda(2);
    complex_double a 
        = sqrt(pow(A(1,1),2)-2.*A(0,0)*A(1,1)+4.*A(0,1)*A(1,0)+pow(A(0,0),2));
    complex_double b = A(0,0)+A(1,1);
    lambda[0] = (b-a)/2.;
    lambda[1] = (b+a)/2.;
    return lambda;
}

std::ostream &operator<<(std::ostream &, const operator_tp &);

/** \relates operator_tp

    Returns the adjungated of \p A (\f$ A^\dagger = (A^T)^* \f$). */
inline operator_tp adj(const operator_tp &A) {
    return
        operator_tp(conj(A(0,0)), conj(A(1,0)),
                    conj(A(0,1)), conj(A(1,1)));
}

/** \relates operator_tp

    Returns the product of \p a and \p B. */
inline operator_tp operator*(complex_double a, const operator_tp &A) {
    return 
        operator_tp(a*A(0,0), a*A(0,1), a*A(1,0), a*A(1,1));
}

/** \relates operator_tp

    Returns the product of \p A and \p B. */
inline operator_tp operator*(const operator_tp &A, const operator_tp &B) {
    return 
        operator_tp(A(0,0)*B(0,0) + A(0,1)*B(1,0),
                    A(0,0)*B(0,1) + A(0,1)*B(1,1),
                    A(1,0)*B(0,0) + A(1,1)*B(1,0),
                    A(1,0)*B(0,1) + A(1,1)*B(1,1));
}

/** \relates operator_tp

    Returns the state that results when \p A is applied to \p psi. */
inline state_tp operator*(const operator_tp &A, const state_tp &psi) {
    return state_tp(A(0,0)*psi(0) + A(0,1)*psi(1),
                    A(1,0)*psi(0) + A(1,1)*psi(1));
}

/** \relates operator_tp

    Returns the absolute square of \p A (\f$ |A|^2 = A A^\dagger\f$). */
inline operator_tp abs_squared(const operator_tp &A) {
    return A*adj(A);
}

/** \relates operator_tp

    Returns the sum of \p A and \p B. */
inline operator_tp operator+(const operator_tp &A, const operator_tp &B) {
    return 
        operator_tp(A(0,0)+B(0,0), A(0,1)+B(0,1),
                    A(1,0)+B(1,0), A(1,1)+B(1,1));
}

/** \relates operator_tp
    
    Adds \p B to \p A and also returns the resulting sum. */
inline const operator_tp &operator+=(operator_tp &A, const operator_tp &B) {
    A(0,0) += B(0,0), A(0,1) += B(0,1), A(1,0) += B(1,0), A(1,1) += B(1,1);
    return A;
}

/** \relates operator_tp

    Returns the result of \p B substracted from \p A. */
inline operator_tp operator-(const operator_tp &A, const operator_tp &B) {
    return 
        operator_tp(A(0,0)-B(0,0), A(0,1)-B(0,1),
                    A(1,0)-B(1,0), A(1,1)-B(1,1));
}

/** \relates operator_tp

    Substracts \p B from \p A and also returns the resulting sum. */
inline const operator_tp &operator-=(operator_tp &A, const operator_tp &B) {
    A(0,0) -= B(0,0), A(0,1) -= B(0,1), A(1,0) -= B(1,0), A(1,1) -= B(1,1);
    return A;
}

/** \relates operator_tp 

    Returns true if \p A is mathematically identical to \p B and false
    otherwise. */
inline bool operator==(const operator_tp &A, const operator_tp &B) {
    return 
        A(0,0)==B(0,0) && A(0,1)==B(0,1) && 
        A(1,0)==B(1,0) && A(1,1)==B(1,1);
}

/** \relates operator_tp
    
    Returns true if \p A is mathematically different from \p B and false
    otherwise. */
inline bool operator!=(const operator_tp &A, const operator_tp &B) {
    return !(A == B);
}

/** \relates operator_tp

    Returns true only if the entries of \p A and \p B are close to each other.

    \sa approx_eq(double, double, approx_tp, double) */
inline bool approx_eq(const operator_tp &A, const operator_tp &B,
                      approx_tp comp = aprx_r, double eps = -1)
{
    return 
        approx_eq(A(0,0), B(0,0), comp, eps) && 
        approx_eq(A(0,1), B(0,1), comp, eps) && 
        approx_eq(A(1,0), B(1,0), comp, eps) && 
        approx_eq(A(1,1), B(1,1), comp, eps);
}

/** \relates operator_tp

    Returns the unitary operator 
        U = exp(-i H t),
    where
        H = E_0 |E_state_0><E_state_0| + E_1 |E_state_1><E_state_1|,
        <E_state_0|E_state_1> = 0,
        dim(E_0) = h_bar/dim(t), dim(E_1) = h_bar/dim(t). */
inline const operator_tp H_to_unitary(double E_0, double E_1, 
                                      const state_tp &E_state_0, 
                                      const state_tp &E_state_1,
                                      double t)
{
    // exp(-i H t) = exp(-i E_0 t) |E_state_0><E_state_0| + 
    //               exp(-i E_1 t) |E_state_1><E_state_1|:
    complex_double i(0,1);
    return exp(-i*E_0*t) * outer_product(E_state_0, E_state_0) +
           exp(-i*E_1*t) * outer_product(E_state_1, E_state_1);
}

state_tp::
state_tp(complex_double psi_0, complex_double psi_1) {
    psi_mb[0] = psi_0, psi_mb[1] = psi_1;
}


state_tp::
state_tp(const complex_double_vect &psi_vect) {
    if (psi_vect.size() != 2)
        throw init_vector_size_error_tp();
    psi_mb[0] = psi_vect[0], psi_mb[1] = psi_vect[1];
}

complex_double &state_tp::operator()(int i) {
    return psi_mb[i];
}

complex_double state_tp::operator()(int i) const {
    return psi_mb[i];
}

/** Returns true if this state is normalized (i.e. its norm is approximately
    one). \p eps specifies the tolerance for numerical comparisons (-1 means to
    use the default value).

    \sa approx_eq(double, double, approx_tp, double) */
bool state_tp::normalized(double eps = -1) const {
    const state_tp &psi = *this;
    return approx_eq(norm(psi(0))+norm(psi(1)), 1., aprx_a, eps);
}

/** Normalizes this state. If this state has length 0 then it is left unchanged. */
void state_tp::normalize() {
    state_tp &psi = *this;
    double length = sqrt(norm(psi(0))+norm(psi(1)));
    if (length != 0.)
        psi(0) /= length, psi(1) /= length;
}

std::ostream &operator<<(std::ostream &, const state_tp &);

/** \relates state_tp

    Returns true only if the entries of \p psi and \p phi are close to each
    other.

    \sa approx_eq(double, double, approx_tp, double) */
inline bool approx_eq(const state_tp &psi, const state_tp &phi,
                      approx_tp comp = aprx_r, double eps = -1)
{
    return 
        approx_eq(psi(0), phi(0), comp, eps) &&
        approx_eq(psi(1), phi(1), comp, eps);
}

/** \relates state_tp

    Returns the state \p psi multiplied by \p x. */
inline state_tp operator*(complex_double x, const state_tp &psi) {
    return state_tp(x*psi(0), x*psi(1));
}

/** \relates state_tp

    Returns the inner product of \p psi and \p phi (\f$psi^\dagger phi\f$). */
inline complex_double operator*(const state_tp &psi, const state_tp &phi) {
    return conj(psi(0))*phi(0) + conj(psi(1))*phi(1);
}

/** \relates state_tp

    Returns the inner product of \p psi and \p phi (\f$psi phi^\dagger\f$). */
inline operator_tp outer_product(const state_tp &psi, const state_tp &phi) {
    complex_double A[2][2];
    for (int i = 0; i <= 1; ++i)
        for (int j = 0; j <= 1; ++j)
            A[i][j] = psi(i) * conj(phi(j));
    return operator_tp(A[0][0], A[0][1], A[1][0], A[1][1]);
}

/** \relates state_tp 

    Returns true if \p psi is mathematically identical to \p phi and false
    otherwise. */
inline bool operator==(const state_tp &psi, const state_tp &phi) {
    return psi(0)==phi(0) && psi(1)==phi(1);
}

/** \relates state_tp
    
    Returns true if \p phi is mathematically different from \p psi and false
    otherwise. */
inline bool operator!=(const state_tp &phi, const state_tp &psi) {
    return !(phi == psi);
}

#endif
