#include "measurement.hpp"
#include "h2.hpp"
#include "misc.hpp"
#include <stdexcept>

using namespace std;

measurement_tp::measurement_tp(const vector<operator_tp> &M, 
                               double completeness_eps) 
    : M_mb(M)
{
    update_effects();
    if (!complete(completeness_eps))
        throw incomplete_error_tp();
}

measurement_tp::measurement_tp(params_tp &params, const string &section, 
                               double completeness_eps)
{
    int N = params.get<int>(section+":N");
    if (N <= 0)
        throw N_out_of_range_error_tp(section);
    for (int i = 0; i < N; ++i) {
        string key(section + ":M" + stringify(i));
        complex_double_vect v = params.get_vector<complex_double >(key);
        M_mb.push_back(operator_tp(v));
    }

    update_effects();
    if (!complete(completeness_eps))
        throw incomplete_error_tp();
}

/** Measures \p psi (thereby changing it) and returns the outcome of the
    measurment. The outcome is a value out of {0, 1, ..., N-1}, where N is the
    number of measurement operators. */
int measurement_tp::measure(state_tp &psi) {
    // Calculation of probabilities p:
    double_vect p;
    p.reserve(M_mb.size());
    for (vector<operator_tp>::const_iterator i = E_mb.begin(); 
         i != E_mb.end(); ++i)
        p.push_back(real(i->expectation_value(psi)));

    int outcome = pick_randomly(p); // Actual measurement with RNG

    // Post measurement state:
    psi = (1/sqrt(p[outcome])) * (M_mb[outcome] * psi);

    return outcome;
}

/** Simulates a measurement of \p psi (thereby changing it) with outcome \p
    outcome. \p outcome must be an element of {0, 1, ..., N-1}, where N is the
    number of measurement operators. If not, the behavior of this function is
    undefined. Also, \p outcome should not have probability zero because, in
    this case, the postmeasurement state is undefined. */
void measurement_tp::measure(state_tp &psi, int outcome) {
    // Probability of the outcome:
    double p_outcome = real(E_mb[outcome].expectation_value(psi));

    // Post measurement state:
    psi = (1/sqrt(p_outcome)) * (M_mb[outcome] * psi);
}

/** Returns true if the set of measurement operators is complete. \p eps
    specifies the tolerance for numerical comparisons (-1 means to use the
    default value). Otherwise returns false. */
bool measurement_tp::complete(double eps = -1) {
    operator_tp A = operator_tp::zero;
    for (vector<operator_tp>::const_iterator i = E_mb.begin(); 
         i != E_mb.end(); ++i)
        A += *i;
    return approx_eq(A, operator_tp::one, aprx_a, eps);
}

/** Calculates the effect operators and stores them in \p E_mb. */
void measurement_tp::update_effects() {
    E_mb.reserve(M_mb.size());
	for (vector<operator_tp>::const_iterator i = M_mb.begin(); 
         i != M_mb.end(); ++i)
        E_mb.push_back(abs_squared(*i));
}

/** Writes the measurement \p measurement as a set of operators to the stream
    \p o. */ 
ostream &operator<<(ostream &o, const measurement_tp &measurement) {
	o << "{";
    vector<operator_tp> M = measurement.M_mb;
	for (vector<operator_tp>::const_iterator i = M.begin(); 
         i != M.end(); ++i)
		o << *i << (((i+1) != M.end()) ? "," : "}");
	return o;
}

/** Returns true if the measurements \p m1 and \p m2 are identical (i.e. have
    identical measurement operators). Otherwise returns false. */
bool operator==(const measurement_tp &m1, const measurement_tp &m2) {
    if (m1.M_mb.size() != m2.M_mb.size())
        return false;
    vector<operator_tp>::const_iterator j = m2.M_mb.begin();
    for (vector<operator_tp>::const_iterator i = m1.M_mb.begin();
         i != m1.M_mb.end(); ++i, ++j) 
    {
        if (*i != *j)
            return false;
    }
    return true;
}

bool operator!=(const measurement_tp &m1, const measurement_tp &m2) {
    return !(m1 == m2);
}

/** Returns true if the measurement operators of \p m1 and \p m2 are
    approximately equal to each other, false otherwise. 

    \sa approx_eq(double, double, approx_tp, double) */
bool approx_eq(const measurement_tp &m1, const measurement_tp &m2, 
               approx_tp comp, double eps) 
{
    if (m1.M_mb.size() != m2.M_mb.size())
        return false;
    vector<operator_tp>::const_iterator j = m2.M_mb.begin();
    for (vector<operator_tp>::const_iterator i = m1.M_mb.begin();
         i != m1.M_mb.end(); ++i, ++j) 
    {
        if (!approx_eq(*i, *j, comp, eps))
            return false;
    }
    return true;
}
