#ifndef TRASIM__MISC_HPP
#define TRASIM__MISC_HPP

/** \file

    Miscellaneous stuff.

    \author Felix E. Klee (felix.klee@inka.de)
    \date 2004 */

#include "mt19937.hpp"
#include <cmath>
#include <complex>
#include <vector>
#include <string>
#include <stdexcept>

/// Pseudorandom number generator for global use
extern mt19937_tp mt19937;

extern double pi;

typedef std::complex<float> complex_float;
typedef std::vector<complex_float> complex_float_vect;
typedef std::vector<float> float_vect;
typedef std::complex<double> complex_double;
typedef std::vector<complex_double> complex_double_vect;
typedef std::vector<double> double_vect;

template<class tp> inline tp square(tp x) { return x*x; }

/** Converts \p x into a string and returns it. */
template<class tp> inline std::string stringify(tp x) {
    using namespace std;
    ostringstream o;
    if (!(o << x))
        throw runtime_error("Couldn't convert value into string.");
    return o.str();
}

/// Default values for \p eps for relative comparison in approx_eq.
const double default_eps_r = 1e-5;

/// Default values for \p eps for absolute comparison in approx_eq.
const double default_eps_a = 1e-5;

enum approx_tp { aprx_r, aprx_a };

/** Compares the floating point values and \p x and \p y and returns true only
    if they are closer to each other than

    - eps*min(x, y) (\p comp = aprx_r = relative comparison), or
    - eps           (\p comp = aprx_a = absolute comparison).

    If \p eps < 0 then \p default_eps_r respectively \p default_eps_a is used
    instead. */
inline bool 
approx_eq(double x, double y, approx_tp comp = aprx_r, 
          double eps = -1)
{
    using namespace std;
    if (comp == aprx_a) {
        if (eps < 0) 
            eps = default_eps_a;
        return abs(x - y) <= eps;
    } else {
        if (eps < 0) 
            eps = default_eps_r;
        return abs(x - y) <= eps * min(abs(x), abs(y));
    }
}

/** Compares the complex floating point values and \p x and \p y and returns
    true only if |x-y| is smaller than 

    - eps*min(x, y) (\p comp = aprx_r = relative comparison), or
    - eps           (\p comp = aprx_a = absolute comparison).

    If \p eps < 0 then \p default_eps_r respectively \p default_eps_a is used
    instead. */
inline bool 
approx_eq(complex_double x, complex_double y, 
          approx_tp comp = aprx_r, double eps = -1) 
{
    using namespace std;
    // Note that below we compare the |x-y|^2 to (eps*...)^2 because this avoids
    // sqrt operations.
    if (comp == aprx_a) {
        if (eps < 0)
            eps = default_eps_a;
        return norm(x - y) <= square(eps);
    } else {
        if (eps < 0) 
            eps = default_eps_r;
        return norm(x - y) <= square(eps) * min(norm(x), norm(y));
    }
}

int pick_randomly(const double_vect &p);

/** Returns the ceiling of x/2 (i.e. the smallest integer value not less than
    x/2). Note that this function only works with integer types. 
    
    Examples: ceil_half(-2) = -1, ceil_half(-1) = 0, ceil_half(0) = 0,
              ceil_half(1)  = 1,  ceil_half(2)  = 1. */
template<class int_tp> inline int_tp ceil_half(int_tp x) {
    return (x%2 == 0) ? x/2 : (x+1)/2;
}

/** Returns the floor of x/2 (i.e.  the largest integer value not greater than
    x/2). Note that this function only works with integer types. 

    Examples: floor_half(-2) = -1, floor_half(-1) = -1, floor_half(0) = 0,
              floor_half(1) = 0, floor_half(2) = 1. */
template<class int_tp> inline int_tp floor_half(int_tp x) {
    return (x < 0) ? (x-1)/2 : x/2;
}

void read_int_vector(std::string, std::vector<int> &, size_t);

#endif
