#ifndef TRASIM__MPF_HPP
#define TRASIM__MPF_HPP

/** \file

    High precision floating point data types based on the \c mpf_class type from
    the GMP library (GNU multiple precision arithmetic library). Among other
    things, this file is part of an implemention of a high precision complex
    floating point class.

    \author Felix E. Klee (felix.klee@inka.de)
    \date 2004 */

#include <gmpxx.h>
#include <complex>
#include <vector>

typedef mpf_class mpf_tp; // Just to have a nicer name

/** Complex class that is based on a real and imaginary parts of type \c
    mpf_tp. Note that the real and imaginary parts always have the same
    precision.

    You may wonder why the std::complex template from the C++ standard library
    is not used instead. Below you find some of the reasons.

    -# According to paragraph 26.2/2 of the standard: The effect of
       instantiating the template complex for any type other than float, double
       or long double is unspecified.
    -# From section 26.1 (Numeric type requirements) of the standard: The
       complex and valarray components are parameterized by the type of
       information they contain and they manipulate.  A C++ program shall
       instantiate these components only with a type T that satisfies the
       following requirements:\n
       [...]\n
       - If T is a class, its assignment operator, copy and default
         constructors, and destructor shall correspond to each other in the
         following sense: Initialization of raw storage using the default
         constructor, followed by assignment, is semantically equivalent to
         initialization of raw storage using the copy constructor. Destruction
         of an object, followed by initialization of its raw storage using the
         copy constructor, is semantically equivalent to assignment to the
         original object.

    The latter point makes the following code example fail when compiled with
    Intel's C++ compiler 8.0 under i386 LINUX (kernel 2.4). This is best
    explained by looking at the constructor of that particular implementation
    of the std::complex class.
    \code
        mpf_class x(5, 128);
        complex<mpf_class> y(x, x);
        assert(x.get_prec() == real(y).get_prec());
    \endcode

    \author Felix E. Klee (felix.klee@inka.de)
    \date 2004 */
class complex_mpf_tp {
public:
    inline complex_mpf_tp(const complex_mpf_tp &);
    inline complex_mpf_tp(const std::complex<double> &, unsigned long);
    inline complex_mpf_tp(const mpf_tp &, const mpf_tp &);
    inline complex_mpf_tp(const mpf_tp &, const mpf_tp &, unsigned long);
    inline const mpf_tp &real() const;
    inline const mpf_tp &imag() const;
    inline complex_mpf_tp conj() const;
    inline unsigned long get_prec() const;
    inline complex_mpf_tp &operator+=(const complex_mpf_tp &);
    inline complex_mpf_tp &operator=(const complex_mpf_tp &);
    inline complex_mpf_tp &operator=(const mpf_tp &);
private:
    mpf_tp re, im;
    friend complex_mpf_tp 
    operator+(const complex_mpf_tp &x, const complex_mpf_tp &);
    friend complex_mpf_tp 
    operator-(const complex_mpf_tp &x, const complex_mpf_tp &);
    friend const mpf_tp &real(const complex_mpf_tp &);
    friend const mpf_tp &imag(const complex_mpf_tp &);
    friend complex_mpf_tp conj(const complex_mpf_tp &);
};

/** Copy constructor: Constructs a complex number with the same value and
    precision as \p x. */
inline complex_mpf_tp::
complex_mpf_tp(const complex_mpf_tp &x) : re(x.real()), im(x.imag()) {}

/** Constructs a complex number whose value is \p x. The precisions of the real
    part and the imaginary part are set to \p prec. */
inline complex_mpf_tp::
complex_mpf_tp(const std::complex<double> &x, unsigned long prec) :
    re(x.real(), prec), im(x.imag(), prec) {}

/** Constructs a complex number with real part \p re and imaginary part \p
    im. The precision of the real and imageinary part are taken from \p re. */
inline complex_mpf_tp::
complex_mpf_tp(const mpf_tp &re, const mpf_tp &im) :
    re(re, re.get_prec()), im(im, re.get_prec()) {}

/** Constructs a complex number with real part \p re and imaginary part \p
    im. The precision of the real and imageinary part is \p prec. */
inline complex_mpf_tp::
complex_mpf_tp(const mpf_tp &re, const mpf_tp &im, unsigned long prec) :
    re(re, prec), im(im, prec) {}

/** Returns the real part. */
inline const mpf_tp &complex_mpf_tp::real() const { return re; }

/** Returns the imaginary part. */
inline const mpf_tp &complex_mpf_tp::imag() const { return im; }

/** Returns the complex conjugate of \c *this. The precision of the conjugate is
    the same as that of \c *this. */
inline complex_mpf_tp complex_mpf_tp::conj() const { 
    return complex_mpf_tp(real(), -imag(), get_prec());
}

/** Returns the precision of the real part and the imaginary part (they are the
    same). */
inline unsigned long complex_mpf_tp::get_prec() const { return re.get_prec(); }

/** Adds \p x to \p *this and returns \p *this. */
inline complex_mpf_tp &complex_mpf_tp::operator+=(const complex_mpf_tp &x) {
    re += x.real();
    im += x.imag();
    return *this;
}

/** Assigns \p x to \p *this and returns \p *this. The precision of \p *this is
    not changed. */
inline complex_mpf_tp &complex_mpf_tp::operator=(const complex_mpf_tp &x) {
    re = x.real();
    im = x.imag();
    return *this;
}

/** Assigns \p x to \p *this and returns \p *this. The precision of \p *this is
    not changed. */
inline complex_mpf_tp &complex_mpf_tp::operator=(const mpf_tp &x) {
    re = x;
    im = 0;
    return *this;
}

/** \relates complex_mpf_tp

    Returns the sum of \p x and \p y. The precision of the result is that of \p
    x. */
inline complex_mpf_tp 
operator+(const complex_mpf_tp &x, const complex_mpf_tp &y) { 
    return complex_mpf_tp(x.real()+y.real(), x.imag()+y.imag(), 
                          x.get_prec());
}

/** \relates complex_mpf_tp

    Returns the sum of \p x and \p -y. The precision of the result is that of
    \p x. */
inline complex_mpf_tp 
operator-(const complex_mpf_tp &x, const complex_mpf_tp &y) { 
    return complex_mpf_tp(x.real()-y.real(), x.imag()-y.imag(), 
                          x.get_prec());
}

/** \relates complex_mpf_tp

    Returns the product of \p x and \p y. The precision of the result is that
    of \p x. */
inline complex_mpf_tp 
operator*(const complex_mpf_tp &x, const complex_mpf_tp &y) { 
    return complex_mpf_tp(x.real()*y.real() - x.imag()*y.imag(),
                          x.real()*y.imag() + x.imag()*y.real(),
                          x.get_prec());
}

/** \relates complex_mpf_tp

    Returns \c x.real(). */
inline const mpf_tp &real(const complex_mpf_tp &x) {
    return x.real();
}

/** \relates complex_mpf_tp

    Returns \c x.imag(). */
inline const mpf_tp &imag(const complex_mpf_tp &x) {
    return x.imag();
}

/** \relates complex_mpf_tp

    Returns \c x.conj(). */
inline complex_mpf_tp conj(const complex_mpf_tp &x) {
    return x.conj();
}

typedef std::vector<complex_mpf_tp> complex_mpf_vect_tp;
typedef std::vector<mpf_tp> mpf_vect_tp;

#endif
