#ifndef TRASIM__PARAMS_HPP
#define TRASIM__PARAMS_HPP

extern "C" {
#include <iniparser.h>
}
#include <string>
#include <sstream>
#include <vector>
#include <stdexcept>

/** An interface for reading and writing data files in ini format (this format
    is popular in the "Windows world"). This is a wrapper for the Iniparser C
    library.

    \sa See also the <a
    href=../3rdparty/iniparser-2.14/html/index.html>Iniparser
    Documentation</a>.
    QQQ: make this local?

    \author Nicolas Devillard <ndevilla@free.fr> (Iniparser)
    \author Felix E. Klee (felix.klee@inka.de) (This interface)
    \date 2004 */
class params_tp {
public:
    params_tp(const std::string &file_name);
    ~params_tp();
    template<class tp> tp get(const std::string &) const;
    template<class tp> tp get(const std::string &, tp) const;
    template<class tp> std::vector<tp> get_vector(const std::string &) const;
    template<class tp> std::vector<tp> get_vector(const std::string &, 
                                                  const std::vector<tp> &) const;
    class convert_error_tp;
    class open_file_error_tp;
    class get_key_error_tp;
private:
    dictionary *ini_dict;
};

class params_tp::convert_error_tp : public std::runtime_error {
 public:
    convert_error_tp(const std::string &s) : 
        runtime_error("Cannot convert \""+s+"\" to the requested type.") {}
};

class params_tp::open_file_error_tp : public std::runtime_error {
 public:
    open_file_error_tp(const std::string &f) :
        runtime_error("Cannot open file \""+f+"\".") {}
};

class params_tp::get_key_error_tp : public std::runtime_error {
 public:
    get_key_error_tp(const std::string &key) :
        runtime_error("Cannot get the string associated with the key \""+
                      key+"\".") {}
};

/** Reads the value of the entry at key \p key, converts it to type \p tp, and
    returns the resulting value. If the key cannot be found or if the value
    cannot be converted to type \p tp, then an exception is thrown. */
template<class tp> tp params_tp::
get(const std::string &key) const {
    using namespace std;
    string s = get<string>(key);
    istringstream sstr(s);
    tp x;
    if (! (sstr >> x))
        throw convert_error_tp(s);
    return x;
}

/** Reads the value of the entry at key \p key, converts it to type \p tp, and
    returns the resulting value. If the key cannot be found or if the value
    cannot be converted to type \p tp, then \p default_value us returned. */
template<class tp> tp params_tp::
get(const std::string &key, tp default_value) const {
    using namespace std;
    try {
        return get<tp>(key);
    } catch(...) {
        return default_value;
    }
}

template<> std::string params_tp::get<std::string>(const std::string &) const;

template<> bool params_tp::get<bool>(const std::string &) const;

/** Reads a vector of values at the key \p key, converts the values to type \p
    tp, and returns the vector. If an error occurs, an exception is thrown. A
    vector might be specified in the data file as follows:

    sample_vector = 1 2 3 4 */
template<class tp> std::vector<tp> params_tp::
get_vector(const std::string &key) const {
    using namespace std;
    string s = get<string>(key);
    istringstream sstr(s);
    vector<tp> v;
    tp x;
    while (sstr >> x)
        v.push_back(x);
    return v;
}

/** Reads a vector of values at the key \p key, converts the values to type \p
    tp, and returns the vector. If an error occurs, \p default_vector is
    returned. A vector might be specified in the data file as follows:

    sample_vector = 1 2 3 4 */
template<class tp> std::vector<tp> params_tp::
get_vector(const std::string &key, const std::vector<tp> &default_vector) 
    const 
{
    using namespace std;
    try {
        return get<vector<tp> >(key);
    } catch(...) {
        return default_vector;
    }
}

#endif
