#include "tracking.hpp"
#include <complex>
using namespace std;

/** \p misc_stream and \p rcrd_stream are streams where miscellaneous and
    measurement related data should be written to. \p section_name is the 
    name of the section in the parameter data params that should is to be 
    used when reading data. */
tracking_tp::tracking_tp(params_tp &params,
						 ostream &misc_stream,
						 ostream &rcrd_stream,
                         string section_name) :
	misc_stream_mb(misc_stream), rcrd_stream_mb(rcrd_stream),
    sec_name_mb(section_name),
    psi_init_mb(params.get_vector<complex<double> >(sec_name_mb+":psi_init")),
    psi_mb(psi_init_mb), psi_tilde_mb(psi_init_mb), 
    t_mb(0),
    E_state_0_mb(params.get_vector
                 <complex<double> >(sec_name_mb+":E_state_0")),
    E_state_1_mb(params.get_vector
                 <complex<double> >(sec_name_mb+":E_state_1")),
    E_0_mb(params.get<double>(sec_name_mb+":E_0")),
    E_1_mb(params.get<double>(sec_name_mb+":E_1")),
	one_mb(0, 1)
{
    if (!psi_init_mb.normalized(1e-4))
        throw psi_init_normalization_error_tp();

    // Checks that E_state_0 and E_state_1 are valid states (i.e. normalized)
    // and orthogonal:
    if (!E_state_0_mb.normalized() ||
        !E_state_1_mb.normalized() ||
        !approx_eq(E_state_0_mb*E_state_1_mb, 0, aprx_a))
        throw E_state_j_orthonormality_error_tp();

    // Initializes the record stream (dim(tracking:E_0)/h_bar is used for the
    // dimension of time where dim(tracking:E_0) is the dimension of the first
    // eigenvector of the Hamiltonian):
    rcrd_stream_mb 
        << "# t/(dim(tracking:E_0)/h_bar) |c_1_tilde|^2 |c_1|^2 g_1\n";
}

tracking_tp::~tracking_tp() {}

/** Records data about the state of the system and the guess g_1. Also streams
    the data to a stream that holds all recordings. Called at each recording
    point.*/
void tracking_tp::record() {
    record_tp rec;
    rec.t_mb = t_mb;
    rec.c_1_tilde_absq_mb = norm(psi_tilde_mb(1));
    rec.c_1_absq_mb = norm(psi_mb(1));
    rec.g_1_mb = g_1_mb;
    recording_mb.push_back(rec);
    stream_rcrd(rec); // Writes information to stream
}

/** Appends measurement steps to the output stream \p rcrd_stream. The
    resulting data consists of lines of the form
    
        t c_1_absq_tilde c_absq_1 g_1,

    where t = "current time", c_1_absq_tilde = |c_1^tilde|^2, c_absq_1 =
    |c_1|^2, g_1 = "current guess" for each measurement step (QQQ: no! - fix
    docu everywhere). */
void tracking_tp::stream_rcrd(const record_tp &rec) {
    rcrd_stream_mb << rec.t_mb << "\t"
                   << rec.c_1_tilde_absq_mb << "\t" 
                   << rec.c_1_absq_mb << "\t" 
                   << rec.g_1_mb;
#if 0 // QQQ: make this configurable or depending on the tracking method
    rcrd_stream_mb << "\n";
#else
    rcrd_stream_mb << endl;
#endif
}
