#ifndef TRASIM__TRACKING_HPP
#define TRASIM__TRACKING_HPP

#include "params.hpp"
#include "h2.hpp"
#include <fstream>

/** This is the superclass of all tracking procedures. 

    \author Felix E. Klee (felix.klee@inka.de)
    \date 2004 */
class tracking_tp {
public:
    struct record_tp;
    class E_state_j_orthonormality_error_tp;
    class psi_init_normalization_error_tp;
    class not_enough_memory_error_tp;
	tracking_tp(params_tp &params, std::ostream &, std::ostream &,
                std::string = "tracking");
	virtual ~tracking_tp();
    virtual unsigned long 
    track() = 0; ///< Simulates the tracking procedure. Returns the
                 ///< number of measurements performed.
    const state_tp &psi_init() const;
    const state_tp &E_state_0() const;
    const state_tp &E_state_1() const;
    double E_0() const;
    double E_1() const;
    const std::vector<record_tp> &recording() const;
protected:
    void record();
    
    void stream_rcrd(const record_tp &rec);
    std::ostream &misc_stream_mb; ///< Stream for misc. data
    std::ostream &rcrd_stream_mb; ///< Stream for measurement related data
    const std::string sec_name_mb; ///< Name of the corresponding section in
                                   ///< the data file
    state_tp psi_init_mb; ///< Initial state
    state_tp psi_mb; ///< Current state |psi(t)>
    state_tp psi_tilde_mb; ///< Current undisturbed state |psi^tilde(t)>
    double t_mb; ///< Current time
    double g_1_mb; // Current guess for |<1|psi(t)>|^2
    state_tp E_state_0_mb, E_state_1_mb; ///< Orthogonal eigenstates of the
                                         ///< Hamiltonian
    double E_0_mb, E_1_mb; ///< Eigenvalues of the Hamiltonian dim(E[j]) =
                           ///< arbitrary units
    std::vector<record_tp> ///< Recorded basic tracking data (for efficiency
    recording_mb;          ///< purposes enough memory for this vector should
                           ///< be reserve()'d before adding elements)
    state_tp one_mb; ///< |1>
};

struct tracking_tp::record_tp {
    double t_mb; ///< Current time (tracking starts always at t_mb = 0)
                 ///< (dim(t_mb = dim(tracking_tp::E_0_mb)/h_bar)
    double c_1_tilde_absq_mb; ///< Value for undisturbed |<1|psi_tilde(t)>|^2
                              ///< (absq = "absolute squared")
    double c_1_absq_mb; ///< Value for disturbed |<1|psi(t)>|^2
    double g_1_mb; ///< Guess g_1(t)
};

class tracking_tp::E_state_j_orthonormality_error_tp : 
    public std::runtime_error 
{
public:
    E_state_j_orthonormality_error_tp() : 
        runtime_error("E_state_0 or E_state_1 is not orthonormal.") {}
};

class tracking_tp::psi_init_normalization_error_tp : 
    public std::runtime_error 
{
public:
    psi_init_normalization_error_tp() : 
        runtime_error("psi_init is not normalized.") {}
};

class tracking_tp::not_enough_memory_error_tp : 
    public std::runtime_error 
{
public:
    not_enough_memory_error_tp() : runtime_error("Not enough memory.") {}
};

/** Returns the the initial state. */
inline const state_tp &tracking_tp::psi_init() const {
    return psi_init_mb;
}

inline const state_tp &tracking_tp::E_state_0() const { return E_state_0_mb; }
inline const state_tp &tracking_tp::E_state_1() const { return E_state_1_mb; }
inline double tracking_tp::E_0() const { return E_0_mb; }
inline double tracking_tp::E_1() const { return E_1_mb; }
inline const std::vector<tracking_tp::record_tp> &tracking_tp::recording() const {
    return recording_mb; 
}

#endif
