#include "trasim.hpp"
#include "tracking.hpp"
#include "uniform-tracking.hpp"
#include "frequency-tracking.hpp"
#include "params.hpp"
#include <cerrno>
#include <iostream>
#include <fstream>
using namespace std;

/** Reads the command line specified by \p argc and \p argv and initializes the
    simulation described in the input file. */
trasim_tp::
trasim_tp(int argc, char **argv) : t(0) {
    if (argc != 2)
        throw cmd_line_error_tp(argv[0]);

    params_tp params(argv[1]);

    misc_file_name_mb = params.get<string>("trasim:misc-file-name");
    misc_stream_mb.open(misc_file_name_mb.c_str());
    if (!misc_stream_mb)
        throw open_file_error_tp(misc_file_name_mb);
    misc_stream_mb << "# -*- mode: ini -*- \n"; // To indicate that the EMACS
                                                // editor should use ini-mode
                                                // when available for viewing
                                                // the file

    rcrd_file_name_mb = params.get<string>("trasim:rcrd-file-name");
    rcrd_stream_mb.open(rcrd_file_name_mb.c_str());
    if (!rcrd_stream_mb)
        throw open_file_error_tp(rcrd_file_name_mb);

    load_random_seed(params);

    string procedure = params.get<string>("trasim:procedure");
    if (procedure == "frequency-tracking")
        t = new frequency_tracking_tp(params, misc_stream_mb, rcrd_stream_mb);
    else if (procedure == "uniform-tracking")
        t = new uniform_tracking_tp(params, misc_stream_mb, rcrd_stream_mb);
    else
        throw procedure_error_tp(procedure);
}

trasim_tp::~trasim_tp() {
    delete t;
}

/** Loads the seed for the RNG from \c trasim:random_seed in \p params and
    initializes the RNG with it. If the seed is invalid an exception of type
    \c invalid_seed_tp is thrown. */
void trasim_tp::load_random_seed(params_tp &params) {
    string s = params.get<string>("trasim:random-seed", "");
    if (s == "time") {
        random_seed_mb = time(0); // Uses time as random seed
    } else if (s == "") {
        random_seed_mb = 0; // Nothing specified => Uses 0 as random seed
    } else if (s[0] == '-') {
        // A negative seed is not allowed. Note that strtoul, used below, does
        // allow negative seeds. Thus, this extra check is necessary.
        throw invalid_seed_error_tp(s); 
    } else {
        char *remainder; // Not converted remainder, e.g. ".2" if seed == "1.2"
        errno = 0;
        random_seed_mb = strtoul(s.c_str(), &remainder, 10);
        if (string(remainder).size() != 0 || errno == ERANGE)
            throw invalid_seed_error_tp(s); // trasim:random_seed contains junk
    }
    mt19937.init_genrand(random_seed_mb);
}

/** Runs the selected tracking procedure and dumps the resulting data to
    files. Also measures the duration of the simulation in seconds. */
void trasim_tp::
run() {
    time_t start_time = time(0);
    t->track();
    time_t end_time = time(0);
    misc_stream_mb << "\n\n";
    misc_stream_mb << "[trasim]\n";
    misc_stream_mb << "tracking-duration = " << end_time-start_time << "\n";
    misc_stream_mb << "random-seed = " << random_seed_mb << "\n";

    // Flushes buffers for the case that they are read shortly after this
    // function terminates:
    misc_stream_mb << flush;
    rcrd_stream_mb << flush;
}
