#include "uniform-tracking.hpp"
#include "uniform-tracking_guess.hpp"
#include "uniform-tracking_const-guess.hpp"
#include "uniform-tracking_aks-guess.hpp"
#include "uniform-tracking_min-dev-nser-guess.hpp"
#include "uniform-tracking_min-dev-xrot-guess.hpp"
#include "uniform-tracking_min-dev-proj-guess.hpp"
#include <ios>
#include <cfloat>
using namespace std;

/** Initializes the uniform tracking procedure and selects the guess. \p params
    provides an interface to the parameters (usually a parameter file). */
uniform_tracking_tp::uniform_tracking_tp(params_tp &params, 
                                         std::ostream &misc_stream,
                                         std::ostream &rcrd_stream) :
	tracking_tp(params, misc_stream, rcrd_stream), // QQQ: pass "uniform-tracking" as section
    U_mb(1,0,0,1),
    M_mb(params.get<int>("uniform-tracking:M")),
    measurement_mb(params, "measurement"),
    guess_mb(0),
    g_1_min_mb(params.get<double>("uniform-tracking:g_1_min",-DBL_MAX)),
    g_1_max_mb(params.get<double>("uniform-tracking:g_1_max",DBL_MAX)),
    n_rec_points_mb(params.get<int>("uniform-tracking:n_rec_points",1)),
    start_delay_mb(params.get<int>("uniform-tracking:start_delay",0)),
    tau_mb(params.get<double>("uniform-tracking:tau")),
    use_predef_outcomes_mb(false)
{
    string s = params.get<string>("uniform-tracking:guess");
    if (s == "constant")
        guess_mb = new const_guess_tp(params);
    else if (s == "aks")
        guess_mb = new aks_guess_tp(params, measurement_mb.M()[0]);
    else if (s == "minimum-deviation-n-series")
        guess_mb = new min_dev_nser_guess_tp(params, measurement_mb.M()[0]);
    else if (s == "minimum-deviation-xrot")
        guess_mb = new min_dev_xrot_guess_tp(params, measurement_mb, M_mb);
    else if (s == "minimum-deviation-projective")
        guess_mb = new min_dev_proj_guess_tp(params, measurement_mb);
    else
        throw wrong_guess_error_tp(s);

    if (n_rec_points_mb < 1)
        // n_rec_points_mb needs to be at least once because otherwise all
        // values of |c_1|^2, all guesses, etc. are lost
        throw n_rec_points_too_small_error_tp();
    
    if (start_delay_mb < 0)
        // Delay cannot be negative
        throw start_delay_too_small_error_tp();
    
    if (tau_mb <= 0)
        throw tau_le_0_error_tp(tau_mb);
    U_mb = H_to_unitary(tracking_tp::E_0(), tracking_tp::E_1(), 
                        tracking_tp::E_state_0(), tracking_tp::E_state_1(), 
                        tau_mb / n_rec_points_mb);

    load_predef_outcomes(params);

    stream_misc();

    try {
        recording_mb.reserve(M_mb * n_rec_points_mb);
    } catch(...) {
        throw not_enough_memory_error_tp();
    }

    try {
        outcomes_mb.reserve(M_mb);
    } catch(...) {
        throw not_enough_memory_error_tp();
    }
}

uniform_tracking_tp::~uniform_tracking_tp() {
    delete guess_mb;
}

const string uniform_tracking_tp::
sec_name() const { return "uniform-tracking"; }

/** If the parameter (read from \p params) \c
    uniform-tracking:predefined-outcomes is non empty, then this function loads
    predefined outcomes from that file. If the file cannot be read (because it
    doesn't exist or due to other errors), then an exception of type \c
    cannot_read_predef_outcomes_error_tp is thrown. If a outcome is out of
    range, then an exception of type \c outcome_out_of_range_error_tp is
    thrown. */
void uniform_tracking_tp::load_predef_outcomes(params_tp &params) {
    string predef_outcomes_filename
        = params.get<string>("uniform-tracking:predefined-outcomes", "");
    if (predef_outcomes_filename != "") {
        // Predefined outcomes should be used => load file
        use_predef_outcomes_mb = true;
        try {
            read_int_vector(predef_outcomes_filename, predef_outcomes_mb, 
                            M_mb);
        } catch (...) {
            throw 
                cannot_read_predef_outcomes_error_tp(predef_outcomes_filename);
        }
        for (unsigned long i = 0; i < M_mb; ++i)
            if (predef_outcomes_mb[i] < 0 || 
                predef_outcomes_mb[i] >= measurement_mb.M().size())
                throw 
                    outcome_out_of_range_error_tp(predef_outcomes_filename, 
                                                  i+1);
    }
}

/** Evolves the system. */
void uniform_tracking_tp::evolve() {
    psi_mb = U_mb * psi_mb;
    psi_tilde_mb = U_mb * psi_tilde_mb;
    
    // Renormalizes psi_mb and psi_tilde_mb to avoid numerical instability:
    psi_mb.normalize();
    psi_tilde_mb.normalize();
}

/** Waits for a time start_delay_mb (measured in recording points) and records
    data at each recording point. */
void uniform_tracking_tp::delay() { // QQQ: put delay in a different class
                                    // (maybe with the following options:
                                    // time_before, time_after, (or use
                                    // inheritance) n_rec_points_before,
                                    // n_rec_points_after)
    g_1_mb = 0;
    for (int i = 0; i < start_delay_mb; ++i) {
        t_mb = - (start_delay_mb - i) * tau_mb / n_rec_points_mb;
        record();
        evolve();
    }
}

/** Waits for a time start_delay_mb (measured in recording points) and then
    simulates the tracking procedure. Returns the number of measurements
    performed. */
unsigned long uniform_tracking_tp::track() {
    psi_tilde_mb = psi_mb = psi_init_mb; // Initial states
    t_mb = 0; // Initial time

    // Writes data concerning the guess to the misc_stream_mb:
    guess_mb->stream_misc(misc_stream_mb);

    delay();

    // Simulates the tracking procedure and records the corresponding data at
    // each recording point:
    unsigned long i;
    t_mb = 0;
    for (i = 0; i < M_mb; ++i) {
        // Performs measurement:
        int outcome;
        if (use_predef_outcomes_mb) {
            outcome = predef_outcomes_mb[i];
            measurement_mb.measure(psi_mb, outcome);
        } else {
            outcome = measurement_mb.measure(psi_mb);
        }
        guess_mb->new_outcome(outcome);
        outcomes_mb.push_back(outcome);
    
        // Records information at the next n_rec_points_mb recording points:
        for (int j = 0; j < n_rec_points_mb; ++j) {
            t_mb = i*tau_mb + j*tau_mb/n_rec_points_mb;
            g_1_mb = guess_mb->guess();
            record();
            if (! (i+1 == M_mb && j+1 == n_rec_points_mb))
                evolve();

            // Stops tracking if guess diverged: (QQQ: move this into
            // tracking_tp?)
            if (g_1_mb < g_1_min_mb || g_1_mb > g_1_max_mb) 
                return i+1; // Number of measurements performed
        }
    }
    return i; // Number of measurements performed
}

/** Appends miscellaneous data to the stream \o misc_stream. */
void uniform_tracking_tp::stream_misc() {
    misc_stream_mb << "\n\n";
    misc_stream_mb << "[uniform-tracking]\n";
    stream_misc_lambda();
}

/** Appends eigenvalues of effect operators and related magnitudes to \p
    misc_steam. */
void uniform_tracking_tp::stream_misc_lambda() {
    misc_stream_mb << "\n";
    misc_stream_mb << "# Eigenvalues of effect operators and related data:\n";

    const vector<operator_tp> &E = measurement_mb.E();
    int j = 0;
    for (vector<operator_tp>::const_iterator i = E.begin(); 
         i != E.end(); ++i, ++j)
    {
        complex_double_vect ev = i->eigenvalues();
        misc_stream_mb << "lambda" << j << "_0 = " << ev[0] << "\n";
        misc_stream_mb << "lambda" << j << "_1 = " << ev[1] << "\n";
        misc_stream_mb << "Delta_lambda" << j << "_abs = " 
                       << abs(ev[1]-ev[0]) << "\n";
        misc_stream_mb << "mean_lambda" << j << "_abs = " 
                       << abs(ev[1]+ev[0])/2. << "\n";
    }
}
