#ifndef TRASIM__UNIFORM_TRACKING_HPP
#define TRASIM__UNIFORM_TRACKING_HPP

#include "params.hpp"
#include "measurement.hpp"
#include "h2.hpp"
#include "tracking.hpp"
#include <stdexcept>
#include <fstream>

/** Uniform Tracking is a tracking procedure where information about the system
    is collected at discrete points in time. The distance in time between the
    measurements is always the same and the measurements are always the same
    (i.e. the measurement operators don't change). That's why we call it
    uniform.

    \author Felix E. Klee (felix.klee@inka.de)
    \date 2004 */
class uniform_tracking_tp : public tracking_tp {
public:
    class wrong_guess_error_tp;
    class tau_le_0_error_tp;
    class n_rec_points_too_small_error_tp;
    class start_delay_too_small_error_tp;
    class cannot_read_predef_outcomes_error_tp;
    class outcome_out_of_range_error_tp;
    class guess_tp;
    class const_guess_tp;
    class min_dev_xrot_guess_tp;
    class nser_guess_tp;
    class min_dev_nser_guess_tp;
    class aks_guess_tp;
    class min_dev_proj_guess_tp;

    uniform_tracking_tp(params_tp &params, std::ostream &, std::ostream &);
    ~uniform_tracking_tp();
    const operator_tp &U() const;
    const measurement_tp &measurement() const;
    const state_tp &psi() const;
    const guess_tp &guess() const;
    virtual unsigned long track();
    int outcome(unsigned long) const;
protected:
    virtual const std::string sec_name() const;
private:
    void load_predef_outcomes(params_tp &);
    void evolve();
    void delay();
    void stream_misc();
    void stream_misc_lambda();

    operator_tp U_mb; ///< Unitary time development
    unsigned long M_mb; ///< Number of measurements (with time evolution in 
                        ///< between) to perform
    measurement_tp measurement_mb; ///< Measurement used
    guess_tp *guess_mb; ///< Guess used
    double g_1_min_mb, g_1_max_mb; ///< A guess outside of this range is
                                   ///< divergent (this causes the tracking
                                   ///< procedure to stop)
    const int 
    n_rec_points_mb; ///< The number of equidistant recording points for each
                     ///< pair of consecutive measurements. A recording point 
                     ///< is a point where information about quantities like
                     ///< e.g. |c_1(t)|^2 is stored. A recording point is 
                     ///< placed immediately after each measurement, and, if
                     ///< n_rec_points_mb>=1, at (n_rec_points_mb-1) points
                     ///< thereafter. Under certain circumstances recording
                     ///< points might also be placed before the tracking
                     ///< procedure starts or after it finishes. From the above
                     ///< explanation it follows that the distance between two
                     ///< points is tau / n_rec_points_mb.
                     ///< 
                     ///< QQQ: put this stuff into a description of the input
                     ///< files?
    const int
    start_delay_mb; ///< Delay after which tracking starts (measured in
                    ///< recording points).
    const double tau_mb; ///< Distance in time between two consecutive
                         ///< measurements (QQQ: dim?)
    bool use_predef_outcomes_mb; ///< True if the measurement outcomes should
                                 ///< be used from the vector \p
                                 ///< predef_outcomes_mb instead of being
                                 ///< generated by the RNG
    std::vector<int> predef_outcomes_mb; ///< Predefined outcomes
    std::vector<int> outcomes_mb; ///< Outcomes of the individual measurements
};               

class uniform_tracking_tp::wrong_guess_error_tp : public std::runtime_error {
public:
    wrong_guess_error_tp(const std::string &s) : 
        runtime_error("Invalid guess \""+s+"\".") {}
};

class uniform_tracking_tp::tau_le_0_error_tp : public std::runtime_error {
public:
    tau_le_0_error_tp(double tau) : 
        runtime_error("tau = "+stringify(tau)+" is not greater than 0.") {}
};

class uniform_tracking_tp::n_rec_points_too_small_error_tp : public std::runtime_error {
public:
    n_rec_points_too_small_error_tp() : 
        runtime_error("n_rec_points needs to be at least 1.") {}
};

class uniform_tracking_tp::start_delay_too_small_error_tp : public std::runtime_error {
public:
    start_delay_too_small_error_tp() : 
        runtime_error("Start delay cannot be negative.") {}
};

class uniform_tracking_tp::cannot_read_predef_outcomes_error_tp : 
    public std::runtime_error {
public:
    cannot_read_predef_outcomes_error_tp(const std::string &s) : 
        runtime_error("Reading outcomes from \""+s+"\" failed.") {}
};

class uniform_tracking_tp::outcome_out_of_range_error_tp : 
    public std::runtime_error {
public:
    outcome_out_of_range_error_tp(const std::string &s, unsigned long i) : 
        runtime_error("Outcome "+stringify(i)+" read from \""+s+
                      "\" is out of range.") {}
};

/** Returns the unitary time evolution operator between log points (= points
    where |c_1|^2, etc. is logged). */
inline const operator_tp &uniform_tracking_tp::U() const {
    return U_mb;
}

/** Returns the measurement. */
inline const measurement_tp &uniform_tracking_tp::measurement() const {
    return measurement_mb;
}

/** Returns the the pretracking state if tracking hasn't been pursued yet,
    otherwise returns the posttracking state. */
inline const state_tp &uniform_tracking_tp::psi() const {
    return psi_mb;
}

/** Returns the selected guess. */
inline const uniform_tracking_tp::guess_tp &uniform_tracking_tp::
guess() const {
    return *guess_mb;
}

/** Returns the outcome of measurement \p n. If tracking hasn't yet been
    pursued or if \p n is out of range, then the result is undefined. */
inline int uniform_tracking_tp::outcome(unsigned long n) const {
    return outcomes_mb[n];
}

#endif
