#include "uniform-tracking_min-dev-xrot-guess.hpp"
#include <cmath>
#include <stdexcept>
#include <string>
#include <vector>
#include <gmpxx.h>
using namespace std;

const string uniform_tracking_tp::min_dev_xrot_guess_tp::
sec_name() const { return "minimum-deviation-xrot-guess"; }

/** \p max_N is the total number of measurements planned. \p meas is the
    measurement used. */
uniform_tracking_tp::min_dev_xrot_guess_tp::
min_dev_xrot_guess_tp(params_tp &params, const measurement_tp &meas, 
                      long max_N) :
    meas_mb(meas), N_mb(0), N0_mb(-1), a_mb(0), a_old_mb(0), 
    precision_mb(read_precision(params)), mpf_0_mb(0., precision_mb),
    complex_mpf_i_mb(0, 1, precision_mb)
{
    I_sc2_mb.resize(max_N, mpf_0_mb);
    pi2_E_tilde_bar_mb.resize(2, mpf_0_mb);

    int max_N0 = max_N-1;
    a_mb = new a_tp(2*max_N0, precision_mb);
    a_old_mb = new a_tp(2*max_N0, precision_mb);
}
                                       
uniform_tracking_tp::min_dev_xrot_guess_tp::~min_dev_xrot_guess_tp() {
    delete a_old_mb;
    delete a_mb;
}

/** Reads the precision from the parameters, corrects it, and returns it. If
    the precision is negative or zero, then the default precision is
    returned. */
unsigned long uniform_tracking_tp::min_dev_xrot_guess_tp::
read_precision(params_tp &params) {
    string s = params.get<string>(sec_name()+":precision", " ");
    if (s[0] == '-') {
        return mpf_get_default_prec(); // Precision is negative
    } else {
        unsigned long p = params.get<unsigned long>(sec_name()+":precision",
                                                    mpf_get_default_prec());
        if (p == 0)
            return mpf_get_default_prec(); // Precision is zero
        else
            return p; // Precision is OK
    }
}

/** Calculates the current guess according to 0306291317. */
void uniform_tracking_tp::min_dev_xrot_guess_tp::new_outcome(int outcome) {
    N_mb ++, N0_mb ++;

    update_I_sc2();
    update_a(meas_mb.M()[outcome]);
    update_pi2_E_tilde_bar();

    mpf_tp tmp1(pi2_E_tilde_bar_00() + pi2_E_tilde_bar_11());
    mpf_tp tmp2(pi2_E_tilde_bar_11() / tmp1);
    guess_mb = tmp2.get_d();
}

/** Calculates {I_sc2(0,N0-0), I_sc2(1,N0-1), I_sc2(2,N0-2), ...,
    I_sc2(N0,N0-N0)} according to the recusion formulas mentioned in
    0306181630. The result is stored in the vector \c I_sc2_mb. */
void uniform_tracking_tp::min_dev_xrot_guess_tp::update_I_sc2() {
    if (N_mb == 1) {
        // I_sc2(0,0):
        I_sc2_mb[0] = 2*pi;
    } else {
        // I_sc2(0,m+1) = (2m+1)/(2m+2) I_sc2(0,m) with m = N0:
        int m = N0_mb - 1;
        I_sc2_mb[0] = (2*m+1) * I_sc2_mb[0] / (2*m+2);

        // I_sc2(n+1,m-1) = (2n+1)/(2m-1) I_sc2(0,m) with 
        // (n,m) in {(0,N0), (1,N0-1), ..., (N0-1,N0-N0+1)}:
        m = N0_mb;
        for (int n = 0; n <= N0_mb-1; ++n, --m)
            I_sc2_mb[n+1] = (2*n+1) * I_sc2_mb[n] / (2*m-1);
    }
}

/** Calculates a_jkn^(N) recursively according to 0307091130. The parameters \p
    j, \p k, \p n are, of course, the indices. The parameter \p M is the
    measurement operator corresponding to the last measurement. Values of
    a_jkn^(N-1) are read from \p a_old. The resulting value of a_jkn^(N) is
    stored in \p a. */
void uniform_tracking_tp::min_dev_xrot_guess_tp::
update_a(int j, int k, int n, complex_mpf_tp M[2][2], a_tp &a, a_tp &a_old) {
    if (N_mb == 1) {
        complex_mpf_tp tmp(0, 0, precision_mb);
        for (int p = 0; p <= 1; ++p)
            tmp += M[j][p] * conj(M[k][p]);
        a.set(j, k, 0, tmp);
    } else {
        complex_mpf_tp tmp(0, 0, precision_mb);
        for (int p = 0; p <= 1; ++p) {
            for (int q = 0; q <= 1; ++q) {
                if (n >= 2)
                    tmp += M[j][p] * conj(M[k][q]) * a_old(p,q,n-2);
                if (n >= 1 && n <= 2*N0_mb-1)
                    tmp += complex_mpf_i_mb * (M[j][1-p]*conj(M[k][q])-
                                M[j][p]*conj(M[k][1-q])) * a_old(p,q,n-1);
                if (n <= 2*N0_mb-2)
                    tmp += M[j][1-p]*conj(M[k][1-q]) * a_old(p,q,n);
            }
        }
        a.set(j, k, n, tmp);
    }
}

/** Calculates a_jkn^(N) for all possible values of j, k, n (N is fixed)
    according to 0307091130. The result is stored in \c a. */
void uniform_tracking_tp::min_dev_xrot_guess_tp::
update_a(const operator_tp &M_m_N0) {
    // Swap a_old_mb and a_mb so that a_old_mb contains the data of a^(N-1):
    a_tp *tmp = a_old_mb;
    a_old_mb = a_mb;
    a_mb = tmp;

    // Converts M from complex<double> to complex_mpf_tp:
    complex_mpf_tp M[2][2] = 
        {{complex_mpf_tp(real(M_m_N0(0,0)),imag(M_m_N0(0,0)),precision_mb),
          complex_mpf_tp(real(M_m_N0(0,1)),imag(M_m_N0(0,1)),precision_mb)},
         {complex_mpf_tp(real(M_m_N0(1,0)),imag(M_m_N0(1,0)),precision_mb),
          complex_mpf_tp(real(M_m_N0(1,1)),imag(M_m_N0(1,1)),precision_mb)}};

    for (int n = 0; n <= 2l*N0_mb; ++n)
        for (int j = 0; j <= 1; ++j)
            for (int k = 0; k <= j; ++k)
                update_a(j, k, n, M, *a_mb, *a_old_mb);
}

/** Calculates 2*pi*E_tilde_bar_jj according to the summation formula in
    0307091720. */
void uniform_tracking_tp::min_dev_xrot_guess_tp::update_pi2_E_tilde_bar() {
    mpf_tp tmp[2] = {0, 0};
    a_tp &a = *a_mb;
    for (int i = 0; i <= 1; ++i)
        pi2_E_tilde_bar_mb[i] = 0;
    for (int n = 0; n <= N0_mb; ++n)
        for (int i = 0; i <= 1; ++i)
            pi2_E_tilde_bar_mb[i] += I_sc2_mb[n] * real(a(i,i,2*n));
}

/** Writes some info (e.g. the precision used) to the miscellaneous stream. */
void uniform_tracking_tp::min_dev_xrot_guess_tp::
stream_misc(ostream &o) const {
    guess_tp::stream_misc(o);

    // GMP may use a higher precision than specified (for example, if that
    // increases performance):
    mpf_tp tmp(0, precision_mb);
	o << "precision-used = " << tmp.get_prec() << "\n";
}
