#ifndef TRASIM__UNIFORM_TRACKING__MIN_DEV_XROT_GUESS_HPP
#define TRASIM__UNIFORM_TRACKING__MIN_DEV_XROT_GUESS_HPP

#include "uniform-tracking_guess.hpp"
#include "misc.hpp"
#include "mpf.hpp"
#include <vector>
#include <stdexcept>
#include <cassert>

/** Minimum deviation guess for observing a qubit that performs a Rabi
    oscillation according to the following Hamiltonian:

    H = h_bar Omega_R/2 (|0><1|+|1><0|),

    where Omega_R is unknown. The guess is calculated in \c new_outcome(int)
    according to 0403060138. One speciality of the guess is that it takes into
    account all previous measurement results. */
class uniform_tracking_tp::min_dev_xrot_guess_tp :
    public uniform_tracking_tp::guess_tp 
{
public:
    class a_tp;
    min_dev_xrot_guess_tp(params_tp &params, const measurement_tp &meas, 
                          long max_N);
	virtual ~min_dev_xrot_guess_tp();
    void new_outcome(int);
    void stream_misc(std::ostream &) const;
    inline const mpf_vect_tp &I_sc2() const;
    inline const a_tp &a() const;
    inline const mpf_tp &pi2_E_tilde_bar_00();
    inline const mpf_tp &pi2_E_tilde_bar_11();
    inline unsigned long precision() const;
protected:
    virtual const std::string sec_name() const;
private:
    unsigned long read_precision(params_tp &);
    void update_I_sc2();
    void update_a(int, int, int, complex_mpf_tp M[2][2], a_tp &, a_tp &);
    void update_a(const operator_tp &M_m_N0);
    void update_pi2_E_tilde_bar();
    const measurement_tp &meas_mb; ///< Measurement used
    long N_mb; ///< Number of measurements so far
    long N0_mb; ///< N_mb - 1
    mpf_vect_tp I_sc2_mb; ///< {I_sc2(0,N0-0), I_sc(1,N0-1),
                          ///<  I_sc2(2,N0-2), ..., I_sc2(N0,N0-N0)}
    a_tp *a_mb;     ///< a_mb(j,k,n) = a_jkn
    a_tp *a_old_mb; ///< Sometimes needed as a backup of a_mb (note that the
                    ///< adresses of a_mb and a_old_mb are sometimes swapped)
    mpf_vect_tp pi2_E_tilde_bar_mb; ///< {2 pi E_tilde_bar_00,
                                    ///<  2 pi E_tilde_bar_11}
    unsigned long precision_mb; ///< Minimum precision (in bits) used for 
                                ///< complex_mpf_tp, etc.. Note that a higher
                                ///< precision might be used if that suits the
                                ///< hardware better (see the GMP 
                                ///< documentation).

    mpf_tp mpf_0_mb; ///< High precision real zero
    complex_mpf_tp complex_mpf_i_mb; ///< High precision complex i
};

/** Container for the values a_jkn^(N) (see 0307091130 for a definition). */
class uniform_tracking_tp::min_dev_xrot_guess_tp::a_tp {
public:
    inline a_tp(int max_n, unsigned long precision);
    inline void set(int, int, int, const complex_mpf_tp &);
    inline complex_mpf_tp operator()(int, int, int) const;
private:
    mpf_vect_tp a_mb; ///< {a_000, a_110, re(a_100), im(a_100), ...}
                      ///< Note that im(a_jjn^(N)) and re(a_01n^(N), 
                      ///< im(a_01n^(N)) are not stored because
                      ///< 1. a_jjn^(N) are real valued (from 0307091630), and
                      ///< 2. a_01n^(N) = (a_10n^(N))^* (also 0307091630).
};

/** Constructs a container for a_jkn^(N) that can hold all values a_jk0, ...,
    a_jk(max_n) (j, k run from 0 to 1). The precision of the values is \p
    precision. */
uniform_tracking_tp::min_dev_xrot_guess_tp::a_tp::
a_tp(int max_n, unsigned long precision) : 
    a_mb(4*(max_n+1), mpf_tp(0.,precision)) {}

/** Stores the value \p x as a_jkn^(N). Note that the combination \p j = 0, \p
    k = 1 is not allowed since it's superfluous: a_01n^(N) can be stored by
    specifying a_10n^(N) = (a_01n^(N))^*. */
void uniform_tracking_tp::min_dev_xrot_guess_tp::a_tp::
set(int j, int k, int n, const complex_mpf_tp &x) {
    using namespace std;
    if (j == 0 && k == 0)
        a_mb[4*n] = real(x);
    else if (j == 1 && k == 1)
        a_mb[4*n+1] = real(x);
    else if (j == 1 && k == 0)
        a_mb[4*n+2] = real(x), a_mb[4*n+3] = imag(x);
    else
        // Shouldn' be reached unless
        // uniform_tracking_tp::min_dev_xrot_guess_tp contains flaws
        assert(false);
}

/** Returns the value a_jkn^(N). */
complex_mpf_tp uniform_tracking_tp::min_dev_xrot_guess_tp::a_tp::
operator()(int j, int k, int n) const {
    if (j == 0 && k == 0)
        return complex_mpf_tp(a_mb[4*n], mpf_tp(0.));
    if (j == 1 && k == 1)
        return complex_mpf_tp(a_mb[4*n+1], mpf_tp(0.));
    if (j == 1 && k == 0)
        return complex_mpf_tp(a_mb[4*n+2], a_mb[4*n+3]);
    else
        return conj(operator()(1,0,n));
}

/** Returns the vector \c I_sc2 which contains {I_sc2(0,N0-0), I_sc(1,N0-1),
    I_sc2(2,N0-2), ..., I_sc2(N0,N0-N0)}. See 0307091720 for a definition
    I_sc2. */
const mpf_vect_tp &uniform_tracking_tp::min_dev_xrot_guess_tp::
I_sc2() const { return I_sc2_mb; }

/** Returns the container of the values a_jkn^(N). See 0307091130 for a
    definition of a_jkn^(N). */
const uniform_tracking_tp::min_dev_xrot_guess_tp::a_tp &
uniform_tracking_tp::min_dev_xrot_guess_tp::
a() const { return *a_mb; }

/** Returns 2*pi*E_tilde_bar_00 (see 0307091720 for a definition). */
const mpf_tp &uniform_tracking_tp::min_dev_xrot_guess_tp::
pi2_E_tilde_bar_00() { return pi2_E_tilde_bar_mb[0]; }

/** Same as \c pi2_E_tilde_bar_00() but with 11 instead of 00. */
const mpf_tp &uniform_tracking_tp::min_dev_xrot_guess_tp::
pi2_E_tilde_bar_11() { return pi2_E_tilde_bar_mb[1]; }

/** Returns the precision (in bits) used for many calculations. Note that a
    higher precision might be used if that suits the hardware better (see the
    GMP (GNU multiple precision arithmetic library - the library used for doing
    high precision arithmetics) documentation. */
unsigned long uniform_tracking_tp::min_dev_xrot_guess_tp::
precision() const { return precision_mb; }

#endif
