#include <measurement.hpp>
#include <h2.hpp>
#include <params.hpp>
#include <frequency-tracking.hpp>
#include "../test-suite/suite.hpp"
#include <iostream>
#include <cmath>
#include <complex>
#include <exception>
#include <cstdlib>
#include <cassert>
#include <cstring>
#include <fstream>
using namespace std;

typedef complex<double> dx;
typedef tracking_tp::record_tp rec_tp;

/** Converts \p x into a string format with precision \p prec. */
static string stringify_with_prec(double x, int prec) {
    ostringstream o;
    o.precision(prec);
    assert(o << x);
    return o.str();
}

/** Some simple tests of error messages, output, etc.. */
static void test1() {
    ostringstream some_stream;

    // E_1 < E_0:
    params_tp params1("test1-1.in");
    try {
        frequency_tracking_tp ft(params1, some_stream, some_stream);
        assert(false);
    } catch (const frequency_tracking_tp::E_1_less_than_E_0_error_tp &e) {
        assert(true);
    }

    // Delta_t <= 0:
    params_tp params2("test1-2.in");
    try {
        frequency_tracking_tp ft(params2, some_stream, some_stream);
        assert(false);
    } catch (const frequency_tracking_tp::Delta_t_too_small_error_tp &e) {
        assert(true);
    }

    // N < 1:
    params_tp params3("test1-3.in");
    try {
        frequency_tracking_tp ft(params3, some_stream, some_stream);
        assert(false);
    } catch (const frequency_tracking_tp::N_too_small_error_tp &e) {
        assert(true);
    }

    // Number of recording points too small:
    params_tp params4("test1-4.in");
    try {
        frequency_tracking_tp ft(params4, some_stream, some_stream);
        assert(false);
    } catch (const frequency_tracking_tp::n_rec_points_too_small_error_tp &e) {
        assert(true);
    }

    // All parameters are OK:
    params_tp params5("test1-5.in");
    try {
        frequency_tracking_tp ft(params5, some_stream, some_stream);
        assert(true);
    } catch (...) {
        assert(false);
    }
}

/* Tests the misc stream before and after tracking. */
void test2() {
    ostringstream rcrd_out;
    ostringstream misc_out;
    misc_out.precision(4);
    string misc_out_exp;

    params_tp params1("test2-1.in");
    frequency_tracking_tp ft1(params1, misc_out, rcrd_out);
    misc_out_exp = 
        "\n\n"
        "[frequency-tracking]\n"
        "\n"
        "omega = 2.537 # Oscillation frequency as defined by E_0 and E_1\n"
        "beta = 0.1224 # Amplitude of the undisturbed osciallation\n"
        "omega_tilde = 2.537 # 0 if beta = 0, omega if beta > 0\n";
    assert(misc_out.str() == misc_out_exp); // Checks misc output before
                                            // tracking
    ft1.track();
    misc_out_exp += "omega_tilde_aprx = " + 
        stringify_with_prec(ft1.omega_tilde_aprx(), 4) + 
        " # Approximated value (by measurements) of omega_tilde\n";
    assert(misc_out.str() == misc_out_exp); // Check misc output after tracking

    // Test misc stream for beta = 0, omega != 0 => omega_tilde = 0:
    params_tp params2("test2-2.in");
    misc_out.str("");
    frequency_tracking_tp ft2(params2, misc_out, rcrd_out);
    misc_out_exp = 
        "\n\n"
        "[frequency-tracking]\n"
        "\n"
        "omega = 5.3 # Oscillation frequency as defined by E_0 and E_1\n"
        "beta = 0 # Amplitude of the undisturbed osciallation\n"
        "omega_tilde = 0 # 0 if beta = 0, omega if beta > 0\n";
    assert(misc_out.str() == misc_out_exp);
    ft2.track();
    misc_out_exp += "omega_tilde_aprx = " + 
        stringify_with_prec(ft2.omega_tilde_aprx(), 4) + 
        " # Approximated value (by measurements) of omega_tilde\n";
    assert(misc_out.str() == misc_out_exp);
}

/** Tests of time evolution and recording of t, c_1_tilde_absq. Note that
    c_1_absq and g_1 are not tested here because they depend on pseudo randomness. */
static void test3() {
    ostringstream dummy_stream;

    params_tp params1("test3-1.in");
    frequency_tracking_tp ft1(params1, dummy_stream, dummy_stream);
    ft1.track();
    const std::vector<tracking_tp::record_tp> &recording = ft1.recording();
#include "test3-1_expected.cpp"
    assert(recording.size() == 18);
    for (size_t i = 0; i < recording.size(); ++i) {
        assert(approx_eq(recording[i].t_mb, 
                         t_expected[i], aprx_a, 1e-4));
        assert(approx_eq(recording[i].c_1_tilde_absq_mb, 
                         c_1_tilde_absq_expected[i], aprx_a, 1e-4));
    }
}

/** Asserts that the elements of \p rec1 in the range \p [ofs1,ofs1+max] are the
    same as the elements of \p rec2 in the range \p [ofs2,ofs2+max]. More
    specifically: \p rec1[ofs1] is compared with \p rec2[ofs2], \p rec2[ofs1+1]
    with \p rec2[ofs2+1], and so on. Note that in the floating point comparisons
    a relatively high tolerance is allowed because the respective ranges of rec1
    and rec2 might have been created under different circumstances. However,
    since, usually, a large number of elements are compared this shouldn't
    diminish the value of this check. */
static void comp_recordings(const vector<rec_tp> &rec1, 
                            const vector<rec_tp> &rec2, 
                            size_t max, size_t ofs1, size_t ofs2) 
{
    const double eps = 5e-2;
    assert(!approx_eq(0.1, 0.2, aprx_a, eps)); // Check that eps is not
                                               // ridiculously large
    for (size_t i = 0; i <= max; ++i) {
        const rec_tp &r1 = rec1[ofs1+i], &r2 = rec2[ofs2+i];
        assert(approx_eq(r1.t_mb, r2.t_mb, aprx_a, eps));
        assert(approx_eq(r1.c_1_tilde_absq_mb, r2.c_1_tilde_absq_mb, aprx_a, eps));
        assert(approx_eq(r1.c_1_absq_mb, r2.c_1_absq_mb, aprx_a, eps));
        assert(approx_eq(r1.g_1_mb, r2.g_1_mb, aprx_a, eps));
    }
}

/** Tests of "frequency-tracking:dont_record". */
static void test4() {
    ostringstream dummy_stream;

    /* Tolerance used for comparing the aprx_omega_tildes. Note that this value
       is somewhat large because not recording data causes - due to numerical
       rounding differences because of differently sized time evolution steps -
       slightly different evolutions of psi_tilde and therefore different
       probabilities for measurement results: */
    double omega_tilde_aprx_eps = 5e-2;

    // Data (dont_record isn't specified => everything should be recorded) that
    // we compare is used as reference data for the various tests below:
    params_tp params("test4.in");
    mt19937.init_genrand(0);
    frequency_tracking_tp ft(params, dummy_stream, dummy_stream);
    ft.track();
    const vector<rec_tp> &rec = ft.recording();
    assert(rec.size() == 90);
    double omega_tilde_aprx = ft.omega_tilde_aprx();
    
    // Test whether not specifying dont_record and specifying for dont_record
    // does the right thing (i.e. everything should be recorded in both cases):
    params_tp params_dr1("test4-dont_record1.in"); // -1 -1 specified
    mt19937.init_genrand(0);
    frequency_tracking_tp ft_dr1(params_dr1, dummy_stream, dummy_stream);
    ft_dr1.track();
    const vector<rec_tp> &rec_dr1 = ft_dr1.recording();
    assert(rec.size() == rec_dr1.size());
    comp_recordings(rec, rec_dr1, rec.size()-1, 0, 0);
    assert(approx_eq(omega_tilde_aprx,          // The approximated omega_tilde
                     ft_dr1.omega_tilde_aprx(), // should be roughly the same
                     aprx_r, omega_tilde_aprx_eps));

    // Test whether specifying (0, N) for the vector dont_record skips all
    // data:
    params_tp params_dr2("test4-dont_record2.in");
    frequency_tracking_tp ft_dr2(params_dr2, dummy_stream, dummy_stream);
    ft_dr2.track();
    assert(ft_dr2.recording().size() == 0);
    assert(approx_eq(omega_tilde_aprx, ft_dr2.omega_tilde_aprx(), aprx_r, 
                     omega_tilde_aprx_eps));

    // Test whether specifying (1) for the vector dont_record skips all
    // data except the data between t=0 and measurement 1:
    params_tp params_dr3("test4-dont_record3.in");
    mt19937.init_genrand(0);
    frequency_tracking_tp ft_dr3(params_dr3, dummy_stream, dummy_stream);
    ft_dr3.track();
    const vector<rec_tp> &rec_dr3 = ft_dr3.recording();
    assert(rec_dr3.size() == 9);
    comp_recordings(rec, rec_dr3, rec_dr3.size()-1, 0, 0);
    assert(approx_eq(omega_tilde_aprx, ft_dr3.omega_tilde_aprx(), aprx_r, 
                     omega_tilde_aprx_eps));

    // Test whether specifying (0 1) for the vector dont_record skips all data
    // between t=0 and measurement 1:
    params_tp params_dr4("test4-dont_record4.in");
    mt19937.init_genrand(0);
    frequency_tracking_tp ft_dr4(params_dr4, dummy_stream, dummy_stream);
    ft_dr4.track();
    const vector<rec_tp> &rec_dr4 = ft_dr4.recording();
    assert(rec_dr4.size() == 90-(1*3+1*6));
    comp_recordings(rec, rec_dr4, rec_dr4.size()-1, 1*3+1*6, 0);
    assert(approx_eq(omega_tilde_aprx, ft_dr4.omega_tilde_aprx(), aprx_r, 
                     omega_tilde_aprx_eps));

    // Test whether specifying (3 5) for the vector dont_record skips all data
    // between measurements 5 and 9:
    params_tp params_dr5("test4-dont_record5.in");
    mt19937.init_genrand(0);
    frequency_tracking_tp ft_dr5(params_dr5, dummy_stream, dummy_stream);
    ft_dr5.track();
    const vector<rec_tp> &rec_dr5 = ft_dr5.recording();
    assert(rec_dr5.size() == 90-(2*3+2*6));
    comp_recordings(rec, rec_dr5, 3*3+3*6-1, 0, 0);
    comp_recordings(rec, rec_dr5, 5*3+5*6-1, 5*3+5*6, 3*3+3*6);
    assert(approx_eq(omega_tilde_aprx, ft_dr5.omega_tilde_aprx(), aprx_r, 
                     omega_tilde_aprx_eps));
}

/** Tests of various parameter combinations. This also involves interactive
    graphical tests where the user has to visually compare results and say
    whether they are roughly similar. */
static void test5() {
    ostringstream misc_stream;

    ofstream rcrd_stream1("test5-1.rcrd");
    assert(rcrd_stream1);
    params_tp params1("test5-1.in");
    frequency_tracking_tp ft1(params1, misc_stream, rcrd_stream1);
    ft1.track();
    // The following test is very rough because on different systems we may get
    // different results because of the combination of system dependent
    // numerical rounding errors and pseudo randomness:
    assert(approx_eq(ft1.omega_tilde(), ft1.omega_tilde_aprx(), aprx_a, 0.5));

    vector<string> gnuplot_files;
    gnuplot_files.push_back("test5-1.gnuplot");
    assert_graphical_outputs_roughly_similar(gnuplot_files);
}

/** Tests of the frequency tracking method. */
int main() {
    vector<test_tp> tests;
    tests.push_back(test1);
    tests.push_back(test2);
    tests.push_back(test3);
    tests.push_back(test4);
    tests.push_back(test5);
    return run_tests(tests, "frequency-tracking");
}
