#include <h2.hpp>
#include <misc.hpp>
#include "../test-suite/suite.hpp"
#include <complex>
#include <iostream>
#include <cassert>
#include <sstream>
#include <stdexcept>
#include <cstdlib>
using namespace std;

typedef complex<double> dx;

/** Test miscellaneous arithmetic and comparison operators. */
static void test1() {
    assert(approx_eq(operator_tp::zero, operator_tp(0,0,0,0), aprx_a));
    assert(approx_eq(operator_tp::one, operator_tp(1,0,0,1), aprx_a));

    operator_tp A(dx(0,1), dx(1,2), dx(2,3), dx(4,5));
    operator_tp B(dx(8,2), dx(9,4), dx(5,8), dx(2,3));

    assert(approx_eq(A(0,0), dx(0,1)) && approx_eq(A(0,1), dx(1,2)) && 
           approx_eq(A(1,0), dx(2,3)) && approx_eq(A(1,1), dx(4,5)));
    assert(approx_eq(B(0,0), dx(8,2)) && approx_eq(B(0,1), dx(9,4)) && 
           approx_eq(B(1,0), dx(5,8)) && approx_eq(B(1,1), dx(2,3)));

    // Test of direct comparison operators:
    assert(A == A);
    assert(!(A == B));
    assert(A != B);

    assert(approx_eq(A+B, operator_tp(dx(8,3), dx(10,6), 
                                      dx(7,11), dx(6,8))));
    assert(approx_eq(A-B, operator_tp(dx(-8,-1), dx(-8,-2), 
                                      dx(-3,-5), dx(2,2))));
    assert(approx_eq(B-A, operator_tp(dx(8,1), dx(8,2), 
                                      dx(3,5), dx(-2,-2))));
    assert(approx_eq(A*B, operator_tp(dx(-13,26), dx(-8,16), 
                                      dx(-10,85), dx(-1,57))));
    assert(approx_eq(B*A, operator_tp(dx(4,43), dx(20,79), 
                                      dx(-13,17), dx(-18,40))));
    assert(approx_eq(adj(A), operator_tp(dx(0,-1), dx(2,-3), 
                                         dx(1,-2), dx(4,-5))));
    assert(approx_eq(adj(B), operator_tp(dx(8,-2), dx(5,-8), 
                                         dx(9,-4), dx(2,-3))));
    assert(approx_eq(abs_squared(A), operator_tp(dx(6,0), dx(17,5), 
                                                 dx(17,-5), dx(54,0))));
    assert(approx_eq(abs_squared(B), operator_tp(dx(165,0), dx(86,-73), 
                                                 dx(86,73), dx(102,0))));

    A += operator_tp::one;
    assert(approx_eq(A, operator_tp(dx(1,1), dx(1,2), 
                                    dx(2,3), dx(5,5))));
    A += B;
    assert(approx_eq(A, operator_tp(dx(9,3), dx(10,6), 
                                    dx(7,11), dx(7,8))));
    A -= operator_tp::one;
    A -= operator_tp::zero;
    A -= B;
    assert(approx_eq(A, operator_tp(dx(0,1), dx(1,2), 
                                    dx(2,3), dx(4,5))));

    ostringstream out;
    out << A;
    assert(out.str() == "(0,1)|0><0|+(1,2)|0><1|+(2,3)|1><0|+(4,5)|1><1|");
    out.str("");
    out << B;
    assert(out.str() == "(8,2)|0><0|+(9,4)|0><1|+(5,8)|1><0|+(2,3)|1><1|");

    vector<dx> A2_vector, B2_vector;
    A2_vector.push_back(dx(0,1)), A2_vector.push_back(dx(1,2));
    A2_vector.push_back(dx(2,3)), A2_vector.push_back(dx(4,5));
    B2_vector.push_back(dx(8,2)), B2_vector.push_back(dx(9,4));
    B2_vector.push_back(dx(5,8)), B2_vector.push_back(dx(2,3));
    operator_tp A2(A2_vector);
    operator_tp B2(B2_vector);
    assert(A == A2);
    assert(B == B2);

    operator_tp A3(A);
    operator_tp B3 = B;
    assert(A3 == A);
    assert(B3 == B);
}

/** Tests initialization of operators with vectors that are too small or too
    large. */
void test2() {
    vector<dx> too_large;
    too_large.push_back(dx(8,2)), too_large.push_back(dx(9,4));
    too_large.push_back(dx(5,8)), too_large.push_back(dx(2,3));
    too_large.push_back(dx(2,9));
    try {
        operator_tp A(too_large);
        assert(false);
    } catch (const operator_tp::init_vector_size_error_tp &e) {
        assert(true);
    }

    vector<dx> too_small;
    too_small.push_back(dx(0,1)), too_small.push_back(dx(1,2));
    too_small.push_back(dx(2,3));
    try {
        operator_tp A(too_small);
        assert(false);
    } catch (const operator_tp::init_vector_size_error_tp &e) {
        assert(true);
    }
}

/** Tests \p approx_eq() for operators. */
void test3() {
    operator_tp A(dx(-3./6321,-7./2), 
                  dx(232./3,5./642), 
                  dx(-763./8,3./465), 
                  dx(4./3,-5./9));
    operator_tp B(A);
    assert(approx_eq(A, B));

    operator_tp 
        A_abs_squared(5482659445021663./914890528818., 
                      dx(1556849178962./15096065040., 5687983701389./15096065040.), 
                      dx(1556849178962./15096065040., -5687983701389./15096065040.), 
                      1133175287809./124545600.);
    assert(approx_eq(abs_squared(A), A_abs_squared));

    operator_tp C = A + operator_tp(0, 0, 0, 1e-3);
    assert(C != A);
    assert(!approx_eq(C, A));
    assert(approx_eq(C, A, aprx_a, 1e-3));
}

/** Tests for state construction and checking. */
static void test4() {
    ostringstream out;

    state_tp psi(dx(-4.5,2.1),dx(-2.3,9.1));
    assert(approx_eq(psi(0), dx(-4.5,2.1)) && approx_eq(psi(1), dx(-2.3,9.1)));
    out << psi;
    assert(out.str() == "(-4.5,2.1)|0>+(-2.3,9.1)|1>");
    assert(!psi.normalized());

    state_tp phi(dx(0.9042,0.3145),dx(0.1671,0.2359));
    assert(approx_eq(phi(0), dx(0.9042,0.3145)) && approx_eq(phi(1), dx(0.1671,0.2359)));
    assert(phi.normalized(1e-4));

    // Same states as above but constructed using vectors:
    vector<dx> psi_vector, phi_vector;
    psi_vector.push_back(dx(-4.5,2.1)), psi_vector.push_back(dx(-2.3,9.1));
    phi_vector.push_back(dx(0.9042,0.3145));
    phi_vector.push_back(dx(0.1671,0.2359));
    state_tp psi2(psi_vector);
    state_tp phi2(phi_vector);
    assert(psi == psi2);
    assert(phi == phi2);

    // Error check if the vector has the wrong size:
    vector<dx> small_vect, large_vect;
    small_vect.push_back(1);
    large_vect.push_back(1), large_vect.push_back(2), large_vect.push_back(3);
    try {
        state_tp psi(small_vect);
        assert(false);
    } catch(const state_tp::init_vector_size_error_tp &e) { 
        assert(true);
    }
    try {
        state_tp psi(large_vect);
        assert(false);
    } catch(const state_tp::init_vector_size_error_tp &e) { 
        assert(true);
    }
}

/** Test of the expectation value function. */
static void test5() {
    operator_tp A(dx(3.5,-12.2),dx(-43.2,5.2),dx(2.4,-43.1),dx(-32.5,-.3));
    state_tp psi(dx(3.4,2.4),dx(5.2,7.3));
    assert(approx_eq(A.expectation_value(psi), dx(-4582.287,-2132.187)));

    assert(approx_eq(operator_tp::one.expectation_value(psi), 97.65));

    assert(approx_eq(operator_tp::zero.expectation_value(psi), 0));
}

/** Test of approx_eq(), operator==(), and operator!=() for states. */
static void test6() {
    state_tp psi(dx(0.4082,0.6804), dx(-0.5443,0.2722));
    state_tp psi_clone(dx(0.4082,0.6804), dx(-0.5443,0.2722));
    state_tp phi(psi(0), psi(1) + 1e-3);

    assert(psi.normalized(1e-4)); // psi is normalized, though this is
                                  // not important here

    assert(approx_eq(psi, psi_clone, aprx_r, 1e-2));
    assert(!approx_eq(psi, phi));
    assert(approx_eq(psi, phi, aprx_r, 1e-2));

    assert(psi == psi_clone);
    assert(psi != phi);
}

/** Test of application of operators to states. */
static void test7() {
    state_tp psi(dx(0.4082,0.6804), dx(-0.5443,0.2722));
    state_tp phi(dx(-0.12,-12.12), dx(23.65,-45.95));
    operator_tp A(dx(3.5,-12.2),dx(-43.2,5.2),dx(2.4,-43.1),dx(-32.5,-.3));
    operator_tp B(dx(5.2,-8.3),dx(-6.3,-123.1),dx(-2.43,0.1),dx(32.4,0.12));

    assert(approx_eq(A*psi, state_tp(dx(31.8279, -17.188), 
                                     dx(48.0763, -24.6437))));
    assert(approx_eq(A*phi, state_tp(dx(-931.024, 2067.06), 
                                     dx(-1305.07, 1462.36))));
    assert(approx_eq(B*psi, state_tp(dx(44.7069, 65.4385),
                                     dx(-18.728, 7.14141))));
    assert(approx_eq(B*phi, state_tp(dx(-5906.66, -2683.86),
                                     dx(773.278, -1456.5))));
}

/** Test of mutliplication of states with scalars. */
static void test8() {
    state_tp psi(dx(0.4082,0.6804), dx(-0.5443,0.2722));
    state_tp phi(dx(-0.12,-12.12), dx(23.65,-45.95));
    double a = -3.954;
    dx b(87.23, -0.12);

    assert(approx_eq(a*psi, state_tp(dx(-1.61402, -2.6903), 
                                     dx(2.15216, -1.07628))));
    assert(approx_eq(b*psi, state_tp(dx(35.6889, 59.3023), 
                                     dx(-47.4466, 23.8093))));
    assert(approx_eq(a*phi, state_tp(dx(0.47448, 47.9225),
                                     dx(-93.5121, 181.686))));
    assert(approx_eq(b*phi, state_tp(dx(-11.922, -1057.21),
                                     dx(2057.48, -4011.06))));
}

/** Test of inner products of states. */
static void test9() {
    state_tp psi(dx(0.4082,0.6804), dx(-0.5443,0.2722));
    state_tp phi(dx(-0.12,-12.12), dx(23.65,-45.95));
    state_tp chi(dx(0.110,0.286), dx(0.864,0.3989));

    assert(approx_eq(psi*phi, dx(-33.6757,13.7073), aprx_a, 1e-4));
    assert(approx_eq(phi*psi, dx(-33.6757,-13.7073), aprx_a, 1e-4));
    assert(approx_eq(psi*chi, dx(-0.122198,-0.410401), aprx_a, 1e-4));
    assert(approx_eq(chi*psi, dx(-0.122198,0.410401), aprx_a, 1e-4));
    assert(approx_eq(phi*chi, dx(-1.37538,50.4337), aprx_a, 1e-4));
    assert(approx_eq(chi*phi, dx(-1.37538,-50.4337), aprx_a, 1e-4));
}

/** Tests \p unitary(). */
static void test10() {
    operator_tp non_U(dx(3.5,-12.2),dx(-43.2,5.2),dx(2.4,-43.1),dx(-32.5,-.3));
    operator_tp U(dx(-0.31558823632335,0.79196093704621), 
                  dx(-0.48555537922359,-0.19348879294225),
                  dx(-0.48555537922359,-0.19348879294225), 
                  dx(-0.31558823632335,0.79196093704621));
    operator_tp approx_U(dx(-0.316,0.792), dx(-0.486,-0.193),
                         dx(-0.486,-0.193), dx(-0.316,0.792));
    assert(!non_U.unitary());
    assert(U.unitary());
    assert(!approx_U.unitary());
    assert(approx_U.unitary(1e-2));
}

/** Tests \p normalize(). */
static void test11() {
    state_tp zero(0,0);
    state_tp zero_expected(0,0);
    assert(!zero.normalized());
    zero.normalize();
    assert(zero == zero_expected);

    state_tp psi(1,2);
    state_tp psi_expected(1./sqrt(5.),2./sqrt(5.));
    assert(!psi.normalized());
    psi.normalize();
    assert(psi.normalized());
    assert(approx_eq(psi, psi_expected));

    state_tp phi(dx(-5.3,2.3),dx(1.5,-9.8));
    state_tp phi_expected(dx(-0.4618835039608, 0.2004400111528),
                          dx(0.130721746404, -0.8540487431728));
    assert(!phi.normalized());
    phi.normalize();
    assert(phi.normalized());
    phi.normalize();
    assert(approx_eq(phi, phi_expected));
}

/** Checks the assertion that \p A.eigenvalues() is approximately equal to \p
    eigenvalues_expected. */
void test12_check(const operator_tp &A,
                  const complex_double_vect &eigenvalues_expected) 
{
    complex_double_vect eigenvalues = A.eigenvalues();
    assert(approx_eq(eigenvalues_expected[0], eigenvalues[0]));
    assert(approx_eq(eigenvalues_expected[1], eigenvalues[1]));
}

/** Tests \p eigenvalues(). */
static void test12() {
    operator_tp A(1, 0, 0, 1);
    complex_double_vect A_ev_expected;
    A_ev_expected.push_back(1);
    A_ev_expected.push_back(1);
    test12_check(A, A_ev_expected);

    operator_tp B(0, 0, 0, 0);
    complex_double_vect B_ev_expected;
    B_ev_expected.push_back(0);
    B_ev_expected.push_back(0);
    test12_check(B, B_ev_expected);

    operator_tp C(dx(3.5,-12.2),dx(-43.2,5.2),dx(2.4,-43.1),dx(-32.5,-.3));
    complex_double_vect C_ev_expected;
    C_ev_expected.push_back(dx(-47.0491,-31.753));
    C_ev_expected.push_back(dx(18.0491,19.253));
    test12_check(C, C_ev_expected);

    operator_tp D(dx(8,2), dx(9,4), dx(5,8), dx(2,3));
    complex_double_vect D_ev_expected;
    D_ev_expected.push_back(dx(-2.52891,-3.41055));
    D_ev_expected.push_back(dx(12.5289,8.41055));
    test12_check(D, D_ev_expected);
}

/** Test of outer products of states. */
static void test13() {
    state_tp psi(dx(0.4082,0.6804), dx(-0.5443,0.2722));
    state_tp phi(dx(-0.12,-12.12), dx(23.65,-45.95));
    state_tp chi(dx(0.110,0.286), dx(0.864,0.3989));

    assert(approx_eq(outer_product(psi, psi), 
                     operator_tp(dx(0.6295714,0.),
                                 dx(-0.036978380000000005,-0.48145375999999995),
                                 dx(-0.036978380000000005,0.48145375999999995),
                                 dx(0.37035533,0.)), aprx_a, 1e-4));
    assert(approx_eq(outer_product(psi, phi), 
                     operator_tp(dx(-8.29543,4.86574),
                                 dx(-21.6105,34.8483),
                                 dx(-3.23375,-6.62958),
                                 dx(-25.3803,-18.5731)), 
                                 aprx_a, 1e-4));
    assert(approx_eq(outer_product(phi, psi), 
                     operator_tp(dx(-8.29543,-4.86574), 
                                 dx(-3.23375,6.62958),
                                 dx(-21.6105,-34.8483),
                                 dx(-25.3803,18.5731)),
                     aprx_a, 1e-4));
    assert(approx_eq(outer_product(psi, chi), 
                     operator_tp(dx(0.239496,-0.0419012), 
                                 dx(0.624096,0.425035),
                                 dx(0.0179762,0.185612),
                                 dx(-0.361695,0.452302)),
                     aprx_a, 1e-4));
    assert(approx_eq(outer_product(chi, psi), 
                     operator_tp(dx(0.239496,0.0419012),
                                 dx(0.0179762,-0.185612),
                                 dx(0.624096,-0.425035),
                                 dx(-0.361695,-0.452302)), 
                     aprx_a, 1e-4));
    assert(approx_eq(outer_product(phi, chi), 
                     operator_tp(dx(-3.47952,-1.29888),
                                 dx(-4.93835,-10.4238),
                                 dx(-10.5402,-11.8184),
                                 dx(2.10414,-49.1348)),
                     aprx_a, 1e-4));
    assert(approx_eq(outer_product(chi, phi),
                     operator_tp(dx(-3.47952,1.29888),
                                 dx(-10.5402,11.8184),
                                 dx(-4.93835,10.4238),
                                 dx(2.10414,49.1348)),
                     aprx_a, 1e-4));
}

/** Test of mutliplication of operators with scalars. */
static void test14() {
    operator_tp A(dx(3.5,-12.2),dx(-43.2,5.2),dx(2.4,-43.1),dx(-32.5,-.3));
    operator_tp B(dx(5.2,-8.3),dx(-6.3,-123.1),dx(-2.43,0.1),dx(32.4,0.12));
    double a = -3.954;
    dx b(87.23, -0.12);

    assert(approx_eq(a*A, 
                     operator_tp(dx(-13.839,48.2388), dx(170.813,-20.5608), 
                                 dx(-9.4896,170.417), dx(128.505,1.1862))));
    assert(approx_eq(b*A, 
                     operator_tp(dx(303.841,-1064.63), dx(-3767.71,458.78),
                                 dx(204.18,-3759.9), dx(-2835.01,-22.269))));
    assert(approx_eq(a*B, 
                     operator_tp(dx(-20.5608,32.8182), dx(24.9102,486.737),
                                 dx(9.60822,-0.3954), dx(-128.11,-0.47448))));
    assert(approx_eq(b*B, 
                     operator_tp(dx(452.6,-724.633), dx(-564.321,-10737.3),
                                 dx(-211.957,9.0146), dx(2826.27,6.5796))));
}

/** Test of the creation of a unitary operator using the function \p
    H_to_unitary. */
static void test15() {
    double E_0, E_1, t;
    state_tp E_0_state(0,0), E_1_state(0,0);

    E_0 = 0.423, E_1 = -1.43;
    E_0_state = state_tp(1,0), E_1_state = state_tp(0,1);
    t = 5.3;
    assert(approx_eq(H_to_unitary(E_0, E_1, E_0_state, E_1_state, t),
                     operator_tp(dx(-0.621851,-0.783136), 0,
                                 0, dx(0.271529,0.96243)), 
                     aprx_a, 1e-4));

    E_0 = -3.12, E_1 = 1.32;
    E_0_state = state_tp(dx(-0.381939,0.354262), dx(0.708524,-0.47604));
    E_1_state = state_tp(dx(0.441196,-0.730731), dx(0.335365,-0.398635));
    t = 1.2;
    assert(approx_eq(H_to_unitary(E_0, E_1, E_0_state, E_1_state, t),
                     operator_tp(dx(-0.23323,-0.882328), 
                                 dx(0.326159,-0.246417),
                                 dx(0.386113,-0.13423), 
                                 dx(-0.603948,-0.684212)), 
                     aprx_a, 1e-4));
}

int main() {
    vector<test_tp> tests;
    tests.push_back(test1);
    tests.push_back(test2);
    tests.push_back(test3);
    tests.push_back(test4);
    tests.push_back(test5);
    tests.push_back(test6);
    tests.push_back(test7);
    tests.push_back(test8);
    tests.push_back(test9);
    tests.push_back(test10);
    tests.push_back(test11);
    tests.push_back(test12);
    tests.push_back(test13);
    tests.push_back(test14);
    tests.push_back(test15);
    return run_tests(tests, "h2");
}
