#include <params.hpp>
#include <measurement.hpp>
#include <h2.hpp>
#include "../test-suite/suite.hpp"
#include <iostream>
#include <cmath>
#include <complex>
#include <cassert>
#include <cstdlib>
using namespace std;

typedef complex<double> dx;

/** Measures N_steps times the state psi with the the measurement. 

    - Compares the post measurement state with the expected states
      psi_post_expected after each measurement.
    - Also compares the probabilities for the outcomes with the expected
      probabilities p_expected. epsilon_post and epsilon_p specify the precision
      for comparing postmeasurement states or probabilities, respectively.
    - Checks that psi stays normalized after n_steps successive measurements. */
void test_measurement(measurement_tp &measurement, 
                      const state_tp &psi, 
                      const vector<double> &p_expected, 
                      const vector<state_tp> &psi_post_expected,
                      double epsilon_post = -1, 
                      long n_steps = 10000,
                      double epsilon_p = 5e-2)
{
    int N = p_expected.size();
    vector<long> p_total(N, 0);

    // Measures:
    for (int i = 0; i < n_steps; ++i) {
        state_tp psi_copy = psi;
        int m = measurement.measure(psi_copy);
        p_total[m] ++;
        assert(approx_eq(psi_copy, psi_post_expected[m], 
                         aprx_a, epsilon_post));
    }

    // Compares probabilities:
    for (int i = 0; i < N; ++i)
        assert(approx_eq(double(p_total[i])/n_steps, p_expected[i], 
                         aprx_a, epsilon_p));

    // Check normalization:
    state_tp psi_copy = psi;
    assert(psi.normalized(1e-4)); // Premeasurement state reasonably normalized?
    for (int i = 0; i < n_steps; ++i)
        measurement.measure(psi_copy);
    assert(psi_copy.normalized());  // Postmeasurement state normalized?
}

/** Verifies that printing "measurement" results in the string str_expected. */
void test_printing(const measurement_tp &measurement, 
                   const string &str_expected) 
{
    ostringstream out;
    out << measurement;
    assert(out.str() == str_expected);
}

/** Loads a measurement from section "section" in file "file_name" and compares
    it to "measurement". */
void test_file_reading(const measurement_tp &measurement, 
                       const string &file_name, const string &section) 
{
    params_tp params(file_name);
    measurement_tp measurement2(params, section);
    assert(measurement2 == measurement);
}

/** Measurement with incomplete measurement operators. => Construction fails. */
static void test1() {
    bool catched;
    vector<operator_tp> Ms;
    Ms.push_back(operator_tp(0.5, 0, 0, 0.5));
    Ms.push_back(operator_tp(0.3, 0.4, 0.6, 0.2));
    try {
        measurement_tp measurement(Ms);
        catched = false;
    } catch (const measurement_tp::incomplete_error_tp &e) {
        catched = true;
    }
    assert(catched);

    // The same measurement when read from a file:
    params_tp params("test1.in");
    try {
        measurement_tp measurement(params, "some-section");
        catched = false;
    } catch (const measurement_tp::incomplete_error_tp &e) {
        catched = true;
    }
    assert(catched);
}

/** Measurement with no measurement operators. => Construction fails because of
    incompleteness. */
void test2() {
    vector<operator_tp> Ms;
    bool catched;
    catched = false;
    try {
        measurement_tp measurement(Ms);
    } catch (const measurement_tp::incomplete_error_tp &e) {
        catched = true;
    }
    assert(catched);

    // The same measurement when read from a file:
    params_tp params("test2.in");
    catched = false;
    try {
        measurement_tp measurement(params, "blabla");
    } catch (const measurement_tp::N_out_of_range_error_tp &e) {
        catched = true;
    }
    assert(catched);
}

/** Measurement with one measurement operator. */
void test3() {
    vector<operator_tp> Ms;
    Ms.push_back(operator_tp::one);
    measurement_tp measurement(Ms);
    test_printing(measurement, "{(1,0)|0><0|+(0,0)|0><1|+(0,0)|1><0|+(1,0)|1><1|}");

    test_file_reading(measurement, "test3.in", "asd");

    // Test of probability distribution of outcomes and test of postmeasurement
    // states:
    state_tp psi = state_tp(dx(0.064451541535, 0.99082369826038), 
                            dx(0.11880284149509, 5.316127149729692e-4));
    vector<double> p_expected;
    p_expected.push_back(1.0);
    vector<state_tp> psi_post_expected;
    psi_post_expected.push_back(psi);
    test_measurement(measurement, psi, p_expected, psi_post_expected, 1e-4);
}

/** Measurement with two measurement operators. */
void test4() {
    // Construction test:
    vector<operator_tp> Ms;
    Ms.push_back(operator_tp(dx(0.85,0.35), dx(-0.05,0.05), 
                             dx(-0.05,0.05), dx(0.85,0.35)));
    Ms.push_back(operator_tp(dx(0.35,0.15), dx(0.05,0.05), 
                             dx(0.05,0.05), dx(0.35,0.15)));
    measurement_tp measurement(Ms);

    test_file_reading(measurement, "test4.in", "gdas");

    // Printing test:
    test_printing(measurement, 
                  "{(0.85,0.35)|0><0|+(-0.05,0.05)|0><1|"
                  "+(-0.05,0.05)|1><0|+(0.85,0.35)|1><1|"
                  ",(0.35,0.15)|0><0|+(0.05,0.05)|0><1|"
                  "+(0.05,0.05)|1><0|+(0.35,0.15)|1><1|}");

    // Test of probability distribution of outcomes and test of postmeasurement
    // states:
    state_tp psi = state_tp(dx(0.9042,0.3145),dx(0.1671,0.2359));
    vector<double> p_expected;
    p_expected.push_back(0.827522);
    p_expected.push_back(0.172537);
    vector<state_tp> psi_post_expected;
    psi_post_expected.push_back(state_tp(dx(0.701723, 0.637975), 
                                         dx(-0.00161045, 0.317127)));
    psi_post_expected.push_back(state_tp(dx(0.640035, 0.640035), 
                                         dx(0.126596, 0.405813)));
    test_measurement(measurement, psi, p_expected, psi_post_expected);
}

/** Measurement with three measurement operators. */
void test5() {
    vector<operator_tp> Ms;
    Ms.clear();
    Ms.push_back(operator_tp(0.3,0,0,0.7));
    Ms.push_back(operator_tp::zero);
    Ms.push_back(operator_tp(sqrt(1-0.3*0.3),0,0,sqrt(1-0.7*0.7)));
    measurement_tp measurement(Ms);

    // Test of probability distribution of outcomes and test of postmeasurement
    // states:
    state_tp psi = state_tp(dx(0.11002678322875,0.28606963639476), 
                            dx(0.86421037008766,0.39899712572681));
    vector<double> p_expected;
    p_expected.push_back(0.452203);
    p_expected.push_back(0);
    p_expected.push_back(0.54731);
    vector<state_tp> psi_post_expected;
    psi_post_expected.push_back(state_tp(dx(0.0490735,0.127591), 
                                         dx(0.899384,0.415236)));
    psi_post_expected.push_back(state_tp(dx(0, 0), 
                                         dx(0, 0)));
    psi_post_expected.push_back(state_tp(dx(0.141839,0.368782),
                                         dx(0.834031,0.385064)));
    test_measurement(measurement, psi, p_expected, psi_post_expected);
}

/** Measurement with three measurement operators. */
void test6() {
    vector<operator_tp> Ms;
    Ms.push_back(operator_tp(0.8, 0, 0, 0.6));
    Ms.push_back(operator_tp(0.6, 0, 0, 0.0));
    Ms.push_back(operator_tp(0.0, 0, 0, 0.8));
    measurement_tp measurement(Ms);

    test_file_reading(measurement, "test6.in", "ioou");

    test_printing(measurement,
                  "{(0.8,0)|0><0|+(0,0)|0><1|+(0,0)|1><0|+(0.6,0)|1><1|,"
                  "(0.6,0)|0><0|+(0,0)|0><1|+(0,0)|1><0|+(0,0)|1><1|,"
                  "(0,0)|0><0|+(0,0)|0><1|+(0,0)|1><0|+(0.8,0)|1><1|}");

    // Test of probability distribution of outcomes and test of postmeasurement
    // states:
    state_tp psi = state_tp(dx(-0.9319,0.004014), dx(0.02110,-0.3621));
    vector<double> p_expected;
    p_expected.push_back(0.603173);
    p_expected.push_back(0.312643);
    p_expected.push_back(0.0841994);
    vector<state_tp> psi_post_expected;
    psi_post_expected.push_back(state_tp(dx(-0.959928, 0.00413472), 
                                         dx(0.016301, -0.279743)));
    psi_post_expected.push_back(state_tp(dx(-0.999991, 0.00430729), 
                                         dx(0, 0)));
    psi_post_expected.push_back(state_tp(dx(0, 0),
                                         dx(0.0581725, -0.998307)));
    test_measurement(measurement, psi, p_expected, psi_post_expected);
}

/** Measurement with two measurement operators. */
void test7() {
    vector<operator_tp> Ms;
    Ms.push_back(operator_tp(dx(0.9,0.3), 0, 0, dx(0.8,0.4)));
    Ms.push_back(operator_tp(dx(0.3,0.1), 0, 0, dx(0.4,0.2)));
    measurement_tp measurement(Ms);

    test_file_reading(measurement, "test7.in", "899s");

    test_printing(measurement,
                  "{(0.9,0.3)|0><0|+(0,0)|0><1|+(0,0)|1><0|+(0.8,0.4)|1><1|,"
                  "(0.3,0.1)|0><0|+(0,0)|0><1|+(0,0)|1><0|+(0.4,0.2)|1><1|}");
}

/** Test of comparison operators (==, !=) and numerical equality operations
    (approx_eq) for measurements. */
void test8() {
    vector<operator_tp> Ms;
    Ms.push_back(operator_tp(dx(0.85,0.35), dx(-0.05,0.05), 
                             dx(-0.05,0.05), dx(0.85,0.35)));
    Ms.push_back(operator_tp(dx(0.35,0.15), dx(0.05,0.05), 
                             dx(0.05,0.05), dx(0.35,0.15)));
    measurement_tp m1(Ms);

    // Identical to m1:
    Ms.clear();
    Ms.push_back(operator_tp(dx(0.85,0.35), dx(-0.05,0.05), 
                             dx(-0.05,0.05), dx(0.85,0.35)));
    Ms.push_back(operator_tp(dx(0.35,0.15), dx(0.05,0.05), 
                             dx(0.05,0.05), dx(0.35,0.15)));
    measurement_tp m2(Ms);

    Ms.clear();
    Ms.push_back(operator_tp::one);
    measurement_tp m3(Ms);

    // Identical to m3:
    Ms.clear();
    Ms.push_back(operator_tp::one);
    measurement_tp m4(Ms);

    Ms.clear();
    Ms.push_back(operator_tp(0.8, 0, 0, 0.6));
    Ms.push_back(operator_tp(0.6, 0, 0, 0.0));
    Ms.push_back(operator_tp(0.0, 0, 0, 0.8));
    measurement_tp m5(Ms);

    // m1 with slight variation:
    Ms.clear();
    Ms.push_back(operator_tp(dx(0.85,0.35), dx(-0.05,0.05), 
                             dx(-0.05,0.05)+dx(3e-6, 2e-6), dx(0.85,0.35)));
    Ms.push_back(operator_tp(dx(0.35,0.15), dx(0.05,0.05), 
                             dx(0.05,0.05), dx(0.35,0.15)+dx(5e-6, 4e-7)));
    measurement_tp m6(Ms, 1e-1);

    // m1 with large variation:
    Ms.clear();
    Ms.push_back(operator_tp(dx(0.85,0.35)+dx(3e-8, 2e-2), dx(-0.05,0.05), 
                             dx(-0.05,0.05), dx(0.85,0.35)+dx(7e-2, 7e-8)));
    Ms.push_back(operator_tp(dx(0.35,0.15), dx(0.05,0.05)+dx(3e-6, 2e-2), 
                             dx(0.05,0.05), dx(0.35,0.15)));
    measurement_tp m7(Ms, 1.);

    assert(m1 == m1);
    assert(m1 == m2);
    assert(m1 != m3);
    assert(m1 != m4);
    assert(m1 != m5);
    assert(m2 == m2);
    assert(m2 != m3);
    assert(m2 != m4);
    assert(m2 != m5);
    assert(m3 == m3);
    assert(m3 == m4);
    assert(m3 != m5);
    assert(m4 == m4);
    assert(m4 != m5);
    assert(m5 == m5);

    assert(approx_eq(m1, m1, aprx_a));
    assert(approx_eq(m1, m2, aprx_a));
    assert(!approx_eq(m1, m3, aprx_a));
    assert(!approx_eq(m1, m4, aprx_a));
    assert(!approx_eq(m1, m5, aprx_a));
    assert(approx_eq(m2, m2, aprx_a));
    assert(!approx_eq(m2, m3, aprx_a));
    assert(!approx_eq(m2, m4, aprx_a));
    assert(!approx_eq(m2, m5, aprx_a));
    assert(approx_eq(m3, m3, aprx_a));
    assert(approx_eq(m3, m4, aprx_a));
    assert(!approx_eq(m3, m5, aprx_a));
    assert(approx_eq(m4, m4, aprx_a));
    assert(!approx_eq(m4, m5, aprx_a));
    assert(approx_eq(m5, m5, aprx_a));
    assert(approx_eq(m1, m6, aprx_a));
    assert(approx_eq(m1, m7, aprx_a, 1e-1));
}

/** Test of reading from broken files. */
void test9() {
    params_tp params("test9.in");

    try {
        measurement_tp m(params, "nonexistent-section-name");
        assert(false);
    } catch (const params_tp::get_key_error_tp &e) {
        assert(true);
    }

    try {
        measurement_tp m(params, "missing-M0");
        assert(false);
    } catch (const params_tp::get_key_error_tp &e) {
        assert(true);
    }

    try {
        measurement_tp m(params, "missing-N");
        assert(false);
    } catch (const params_tp::get_key_error_tp &e) {
        assert(true);
    }

    try {
        measurement_tp m(params, "empty");
        assert(false);
    } catch (const params_tp::get_key_error_tp &e) {
        assert(true);
    }

    try {
        measurement_tp m(params, "nonnumerical-N");
        assert(false);
    } catch (const params_tp::convert_error_tp &e) {
        assert(true);
    }

    try {
        measurement_tp m(params, "negative-N");
        assert(false);
    } catch (const measurement_tp::N_out_of_range_error_tp &e) {
        assert(true);
    }

    try {
        measurement_tp m(params, "incomplete-M0");
        assert(false);
    } catch (const operator_tp::init_vector_size_error_tp &e) {
        assert(true);
    }
}

/** Test of measurements with preset outcomes. */
void test10() {
    state_tp psi(1, 0); // Temporary state

    // Measurement with one measurement operator:
    state_tp psi1_pre = state_tp(dx(0.064451541535, 0.99082369826038), 
                                 dx(0.11880284149509, 5.316127149729692e-4));
    state_tp psi1_post_expected = psi1_pre; // The state should remain unchanged
    vector<operator_tp> Ms1;
    Ms1.push_back(operator_tp(operator_tp::one));
    measurement_tp measurement1(Ms1);
    psi = psi1_pre;
    measurement1.measure(psi, 0);
    assert(approx_eq(psi, psi1_post_expected, aprx_a));

    // Measurement with two measurement operators:
    state_tp psi2_pre = state_tp(dx(0.9042,0.3145),dx(0.1671,0.2359));
    state_tp psi2_post_expected_0 = state_tp(dx(0.701723, 0.637975), 
                                             dx(-0.00161045, 0.317127));
    state_tp psi2_post_expected_1 = state_tp(dx(0.640035, 0.640035), 
                                             dx(0.126596, 0.405813));
    vector<operator_tp> Ms2;
    Ms2.push_back(operator_tp(dx(0.85,0.35), dx(-0.05,0.05), 
                              dx(-0.05,0.05), dx(0.85,0.35)));
    Ms2.push_back(operator_tp(dx(0.35,0.15), dx(0.05,0.05), 
                              dx(0.05,0.05), dx(0.35,0.15)));
    measurement_tp measurement2(Ms2);
    psi = psi2_pre;
    measurement2.measure(psi, 0);
    assert(approx_eq(psi, psi2_post_expected_0, aprx_a));
    psi = psi2_pre;
    measurement2.measure(psi, 1);
    assert(approx_eq(psi, psi2_post_expected_1, aprx_a));

    // Measurement with three measurement operators:
    state_tp psi3_pre = state_tp(dx(0.11002678322875,0.28606963639476), 
                                 dx(0.86421037008766,0.39899712572681));
    state_tp psi3_post_expected_0 = state_tp(dx(0.0490735,0.127591), 
                                             dx(0.899384,0.415236));
    state_tp psi3_post_expected_2 = state_tp(dx(0.141839,0.368782),
                                             dx(0.834031,0.385064));
    vector<operator_tp> Ms3;
    Ms3.push_back(operator_tp(0.3,0,0,0.7));
    Ms3.push_back(operator_tp::zero);
    Ms3.push_back(operator_tp(sqrt(1-0.3*0.3),0,0,sqrt(1-0.7*0.7)));
    measurement_tp measurement3(Ms3);
    psi = psi3_pre;
    measurement3.measure(psi, 0);
    assert(approx_eq(psi, psi3_post_expected_0, aprx_a));
    // Outcome 1 needs to be avoided because M_1 = 0 and, therefore, the
    // corresponding postmeasurement state is undefined.
    psi = psi3_pre;
    measurement3.measure(psi, 2);
    assert(approx_eq(psi, psi3_post_expected_2, aprx_a));
}

int main() {
    vector<test_tp> tests;
    tests.push_back(test1);
    tests.push_back(test2);
    tests.push_back(test3);
    tests.push_back(test4);
    tests.push_back(test5);
    tests.push_back(test6);
    tests.push_back(test7);
    tests.push_back(test8);
    tests.push_back(test9);
    tests.push_back(test10);
    return run_tests(tests, "measurement");
}
