#include <h2.hpp>
#include <misc.hpp>
#include "../test-suite/suite.hpp"
#include <complex>
#include <iostream>
#include <cassert>
#include <sstream>
#include <stdexcept>
#include <cstdlib>
#include <vector>
using namespace std;

typedef complex<double> dx;

/** Tests of pick_randomly(). */
static void test1() {
    mt19937.init_genrand(123);
    
    vector<double> p;
    p.push_back(0.5);
    p.push_back(0.5);
    long total = 0;
    for (int i = 0; i < 10000l; ++i)
        total += pick_randomly(p);
    assert(approx_eq(double(total)/10000l, 0.5, aprx_a, 1e-2));

    p.clear();
    p.push_back(1.);
    total = 0;
    for (int i = 0; i < 10000l; ++i)
        assert(pick_randomly(p) == 0);

    p.clear();
    p.push_back(0.);
    p.push_back(0.);
    p.push_back(1.);
    total = 0;
    for (int i = 0; i < 10000l; ++i)
        total += pick_randomly(p);
    assert(approx_eq(double(total)/10000l, 0.*0+0.*1+1.*2, aprx_a));

    p.clear();
    p.push_back(0.5);
    p.push_back(0.2);
    p.push_back(0.3);
    total = 0;
    for (int i = 0; i < 10000l; ++i)
        total += pick_randomly(p);
    assert(approx_eq(double(total)/10000l, 0.5*0+0.2*1+0.3*2, aprx_a, 0.01));

    // Empty probabilities => pick_randomly() should return -1.
    p.clear();
    total = 0;
    for (int i = 0; i < 10000l; ++i)
        assert(pick_randomly(p) == -1);

    // Sum of probabilities is smaller than 1 => The highest value should be
    // more probable than intended.
    p.clear();
    p.push_back(0.1);
    p.push_back(0.3);
    p.push_back(0.2);
    total = 0;
    for (int i = 0; i < 10000l; ++i)
        total += pick_randomly(p);
    assert(approx_eq(double(total)/10000l, 0.1*0+0.3*1+0.6*2, aprx_a, 0.01));

    // Sum of probabilities is larger than 1 => The highest values should be
    // less probable (or even improbable) than intended.
    p.clear();
    p.push_back(0.2);
    p.push_back(0.5);
    p.push_back(0.2);
    p.push_back(0.7);
    p.push_back(0.4);
    total = 0;
    for (int i = 0; i < 20000l; ++i)
        total += pick_randomly(p);
    assert(approx_eq(double(total)/20000l, 0.2*0+0.5*1+0.2*2+0.1*3+0.0*0, aprx_a, 0.01));
}

/** Tests of stringify(). */
static void test2() {
    assert(stringify(5) == "5");
    assert(stringify(dx(3,4)) == "(3,4)");
    assert(stringify("hello") == "hello");
}

/** Tests of approx_eq_*. */
static void test3() {
    assert(approx_eq(3.5, 3.5));
    assert(approx_eq(3.5, 3.5 + 0.9e-5));
    assert(approx_eq(3.5, 3.5, aprx_a));
    assert(approx_eq(3.5, 3.5 + 0.9e-5, aprx_a));

    assert(!approx_eq(3.5, 3.6));
    assert(approx_eq(3.5, 3.6, aprx_r, 2e-1));
    assert(!approx_eq(3.5, 3.6, aprx_a));
    assert(approx_eq(3.5, 3.6, aprx_a, 2e-1));

    assert(!approx_eq(3.5, -4.5));
    assert(approx_eq(-4.5, -4.5));
    assert(approx_eq(-4.5, -4.5+1e-5));
    assert(!approx_eq(3.5, -4.5, aprx_a));
    assert(approx_eq(-4.5, -4.5, aprx_a));
    assert(approx_eq(-4.5, -4.5+1e-5, aprx_a));

    assert(!approx_eq(0.001, 0, aprx_r, 2));
    assert(approx_eq(0.001, 0, aprx_a, 0.002));

    assert(!approx_eq(0.001, 0, aprx_r, 2));
    assert(approx_eq(0, 0.001, aprx_a, 0.002));

    assert(approx_eq(dx(3.5, -4), dx(3.5, -4)));
    assert(!approx_eq(dx(3.5, -4), dx(-2, 3)));
    assert(approx_eq(dx(3.5, -4+2e-5), dx(3.5, -4)));
    assert(approx_eq(dx(3.5, -4+0.1), dx(3.5, -4), aprx_r, 0.1));
    assert(approx_eq(dx(3.5, -4), dx(3.5, -4), aprx_a));
    assert(!approx_eq(dx(3.5, -4), dx(-2, 3), aprx_a));
    assert(approx_eq(dx(3.5, -4+0.9e-5), dx(3.5, -4), aprx_a));
    assert(approx_eq(dx(3.5, -4+0.1), dx(3.5, -4), aprx_a, 0.2));
}

/** Tests of ceil_half and floor_half. */
static void test4() {
    assert(ceil_half(-4) == -2);
    assert(ceil_half(-3) == -1);
    assert(ceil_half(-2) == -1);
    assert(ceil_half(-1) == 0);
    assert(ceil_half(0) == 0);
    assert(ceil_half(1) == 1);
    assert(ceil_half(2) == 1);
    assert(ceil_half(3) == 2);
    assert(ceil_half(4) == 2);

    assert(floor_half(-4) == -2);
    assert(floor_half(-3) == -2);
    assert(floor_half(-2) == -1);
    assert(floor_half(-1) == -1);
    assert(floor_half(0) == 0);
    assert(floor_half(1) == 0);
    assert(floor_half(2) == 1);
    assert(floor_half(3) == 1);
    assert(floor_half(4) == 2);
}

/** Tests of read_int_vector. */
static void test5() {
    vector<int> v;

    // File doesn't exist:
    try {
        read_int_vector("non-existent-file", v, 10);
        assert(false);
    } catch (const runtime_error &e) {
        assert(string(e.what()) == 
               string("Cannot open file \"non-existent-file\"."));
    }
    
    // File contains less entries than 10:
    try {
        read_int_vector("test5-1.int", v, 10);
        assert(false);
    } catch (const runtime_error &e) {
        assert(string(e.what()) == 
               string("Reading file \"test5-1.int\" failed."));
    }

    // File contains non integer entries among the first 8 entries
    try {
        read_int_vector("test5-2.int", v, 9);
        assert(false);
    } catch (const runtime_error &e) {
        assert(string(e.what()) == 
               string("Reading file \"test5-2.int\" failed."));
    }

    // File is empty and N is 0, reading should work:
    read_int_vector("test5-3.int", v, 0);
    assert(v.size() == 0);

    // Reading should work:
    read_int_vector("test5-1.int", v, 7);
    assert(v.size() == 7 && v[0] == 11 && v[1] == 2 && v[2] == 14 && 
           v[3] == 4 && v[4] == 8 && v[5] == 6 && v[6] == 7);
}

int main() {
    vector<test_tp> tests;
    tests.push_back(test1);
    tests.push_back(test2);
    tests.push_back(test3);
    tests.push_back(test4);
    tests.push_back(test5);
    return run_tests(tests, "misc");
}
