#include <mpf.hpp>
#include <misc.hpp>
#include <../test-suite/suite.hpp>
#include <complex>
#include <iostream>
#include <cassert>
using namespace std;

/** Miscellaneous tests of \t complex_mpf_tp. */
static void test1() {
    // Constructor from type complex<double> and copy constructor and test of
    // get_prec(), real(), and imag():
    complex<double> x1(1.23, -5.345);
    complex_mpf_tp y1(x1, 75);
    complex_mpf_tp z1(y1);
    assert(approx_eq(real(y1).get_d(), real(x1)));
    assert(approx_eq(imag(y1).get_d(), imag(x1)));
    assert(approx_eq(real(y1).get_d(), y1.real().get_d()));
    assert(approx_eq(imag(y1).get_d(), y1.imag().get_d()));
    assert(approx_eq(real(z1).get_d(), real(x1)));
    assert(approx_eq(imag(z1).get_d(), imag(x1)));
    assert(real(y1).get_prec() >= 75);
    assert(imag(y1).get_prec() >= 75);
    assert(y1.get_prec() == real(y1).get_prec());
    assert(real(z1).get_prec() >= 75);
    assert(imag(z1).get_prec() >= 75);
    assert(z1.get_prec() == real(z1).get_prec());

    // Constructor from two mpf_tp numbers and test of get_prec():
    mpf_tp re2(12.75, 256), im2(-543.12, 128);
    complex_mpf_tp x2(re2, im2);
    assert(approx_eq(real(x2).get_d(), 12.75));
    assert(approx_eq(imag(x2).get_d(), -543.12));
    assert(real(x2).get_prec() == re2.get_prec());
    assert(imag(x2).get_prec() == real(x2).get_prec());
    assert(x2.get_prec() == x2.real().get_prec());

    // Constructor from two mpf_tp numbers:
    mpf_tp re3(12.75, 256), im3(-543.12, 128);
    complex_mpf_tp x3(re3, im3, 8192);
    assert(approx_eq(real(x3).get_d(), 12.75));
    assert(approx_eq(imag(x3).get_d(), -543.12));
    assert(real(x3).get_prec() >= 8192);
    assert(imag(x3).get_prec() == real(x3).get_prec());
    assert(x3.get_prec() >= 8192);

    // Complex conjugate conj():
    mpf_tp re4(4.23, 256), im4(-43.2, 128);
    complex_mpf_tp x4(re4, im4, 8192);
    complex_mpf_tp y4(x4.conj());
    complex_mpf_tp z4(conj(x4));
    assert(approx_eq(real(y4).get_d(), 4.23));
    assert(approx_eq(imag(y4).get_d(), 43.2));
    assert(y4.get_prec() >= 8192);
    assert(real(z4).get_d() == real(y4).get_d());
    assert(imag(z4).get_d() == imag(y4).get_d());
    assert(z4.get_prec() == y4.get_prec());

    // Assignment:
    complex_mpf_tp x5(4.32, 0.034, 8192);
    complex_mpf_tp y5(3.43, 1231.23, 128);
    mpf_tp z5(543.12, 64);
    x5 = y5;
    assert(approx_eq(real(x5).get_d(), real(y5).get_d()));
    assert(approx_eq(imag(x5).get_d(), imag(y5).get_d()));
    assert(x5.get_prec() >= 8192);
    x5 = z5;
    assert(approx_eq(real(x5).get_d(), z5.get_d()));
    assert(approx_eq(imag(x5).get_d(), 0.));
    assert(x5.get_prec() >= 8192);
}

/** Tests of multiplying and adding numbers of type \t complex_mpf_tp. */
static void test2() {
    // Sums:
    complex_mpf_tp x1(complex<double>(1.5, -2.3), 128);
    complex_mpf_tp y1(complex<double>(-4.6, 6.7), 256);
    x1 += y1;
    assert(approx_eq(real(x1).get_d(), -3.1));
    assert(approx_eq(imag(x1).get_d(), 4.4));
    x1 += x1;
    assert(approx_eq(real(x1).get_d(), -6.2));
    assert(approx_eq(imag(x1).get_d(), 8.8));
    complex_mpf_tp z1(x1 + y1);
    assert(approx_eq(real(z1).get_d(), -10.8));
    assert(approx_eq(imag(z1).get_d(), 15.5));
    assert(real(z1).get_prec() == real(x1).get_prec());

    // Differences:
    complex_mpf_tp x2(complex<double>(1.5, -2.3), 128);
    complex_mpf_tp y2(complex<double>(-4.6, 6.7), 256);
    complex_mpf_tp z2(x2 - y2);
    assert(approx_eq(real(z2).get_d(), 6.1));
    assert(approx_eq(imag(z2).get_d(), -9.));
    assert(real(z2).get_prec() == real(x2).get_prec());
    z2 = y2 - x2;
    assert(approx_eq(real(z2).get_d(), -6.1));
    assert(approx_eq(imag(z2).get_d(), 9.));
    assert(real(z2).get_prec() == // Assignment doesn't change precision
           real(x2).get_prec());

    // Products:
    complex_mpf_tp x3(complex<double>(0.324, 123.2), 1024);
    complex_mpf_tp y3(complex<double>(-432.23, 0.0024), 256);
    complex_mpf_tp a3(x3*x3);
    complex_mpf_tp b3(x3*y3);
    complex_mpf_tp c3(y3*x3);
    complex_mpf_tp d3(y3*y3);
    assert(approx_eq(real(a3).get_d(), -15178.1));
    assert(approx_eq(imag(a3).get_d(), 79.8336));
    assert(a3.get_prec() >= 1024);
    assert(approx_eq(real(b3).get_d(), -140.338));
    assert(approx_eq(imag(b3).get_d(), -53250.7));
    assert(b3.get_prec() >= 1024);
    assert(approx_eq(real(c3).get_d(), -140.338));
    assert(approx_eq(imag(c3).get_d(), -53250.7));
    assert(c3.get_prec() >= 256);
    assert(approx_eq(real(d3).get_d(), 186823.));
    assert(approx_eq(imag(d3).get_d(), -2.0747));
    assert(d3.get_prec() >= 256);
}

int main() {
    vector<test_tp> tests;
    tests.push_back(test1);
    tests.push_back(test2);
    return run_tests(tests, "mpf");
}
