#include "suite.hpp"
#include <cstdlib>
#include <iostream>
#include <stdexcept>
#include <cstdlib>
#include <fstream>
#include <string>

/** Executes all the tests in the vector \p tests. Returns \p EXIT_SUCCESS if
    all tests passed, otherwise returns EXIT_FAILURE. \p tests_name is the name
    of the unit the tests are related to. */
int run_tests(const vector<test_tp> &tests, const string &tests_name) {
    cout << "Testing " << tests_name << " ... ";
    cout.flush();
    try {
        for (vector<test_tp>::const_iterator i = tests.begin(); 
             i != tests.end(); ++i)
        {
                (*i)();
        }
    } catch(const exception &e) {
        cout << "failed.\n";
        cerr << e.what() << "\n";
        return EXIT_FAILURE;
    }
    cout << "passed.\n";
    return EXIT_SUCCESS;
}

/** Loops until the user enters "y", "n", or something that cannot be converted
    into a string (usually this shouldn't happen). Returns true if the user
    entered "y" and false otherwise. */
bool query_y_n() {
    string q;
    while (std::cin >> q) {
        if (q == "y")
            return true;
        else if (q == "n")
            return false;
    }
    return false; // User entered garbage
}

/** Presents the user with the plots described by the gnuplot input files
    listed in \p gnuplot_files. Each such file should contain instructions for
    a plot that lets the user visually compare the output by a test with the
    expected output. After being presented with the plots the user has to
    decide whether they look roughly similar. For a more detailed description
    see the usage instruction displayed by this function on stdout. */
void assert_graphical_outputs_roughly_similar
(const vector<string> &gnuplot_files)
{
    cout << "\n";
    cout << "We will now try to perform some graphical tests.\n";
    cout << "\n";
    cout << 
        "Do you have GNUPLOT (should be reasonably compatible with version 3.7)\n"
        "installed so that it can be executed by calling \"gnuplot\"? [y|n] ";
    cout.flush();
    if (!query_y_n())
        return;

    cout << 
        "\n"
        "You will now be presented with several gnuplot plots which contain the output\n"
        "of a test and its expected output. Then you will be asked whether the outputs\n"
        "look roughly similar. Note that the outputs are most likely not identical\n"
        "because of the combination of numerical rounding errors and pseudo randomness\n"
        "in the simulation. The outputs are defined to look roughly similar when the\n"
        "corresponding curves have similar features (i.e. the periods and amplitudes of\n"
        "oscillations are similar, etc.) but not neccessarily have the same peaks, etc..\n"
        "\n"
        "Continue? [y|n] ";
    cout.flush();
    if (!query_y_n())
        return;

    for (vector<string>::const_iterator i = gnuplot_files.begin();
         i != gnuplot_files.end(); ++i)
    {
        system(string("gnuplot -persist -raise " + *i).c_str());
    }

    // Performs assertion:
    cout << 
        "\n"
        "Do the outputs of the tests and the expected outputs look roughly\n"
        "similar? [y|n] ";
    cout.flush();
    if (!query_y_n()) { 
        cerr << 
            "\n"
            "Error: Graphical outputs did not look roughly similar." << endl;
        abort();
    }
}

/** Returns line number \p n (counting starts at zero) from the text file named
    \p filename. If there are more than 1024 characters in a line, then this
    function becomes unreliable. */
string getln(string filename, unsigned int n) {
    ifstream file(filename.c_str());
    assert(file);
    int i = 0;
    while (file) {
        char line[1025];
        file.getline(line, 1025);
        if (i++ == n)
            return string(line);
    }
    assert(false);
    return "";
}

/** Writes the string \p s to a file named \p filename. */
void string_to_file(string file_name, const string &s) {
    ofstream out;
    out.open(file_name.c_str());
    assert(out);
    out << s;
}
