#include <trasim.hpp>
#include <params.hpp>
#include "../test-suite/suite.hpp"
#include <cstdlib>
#include <iostream>
#include <cassert>
using namespace std;

/** Tests some error handling. */
static void test1() {
    // Too few parameters:
    try {
        int argc = 1;
        char *argv[] = {"some_app_name"};
        trasim_tp trasim(argc, argv);
        assert(false);
    } catch (const trasim_tp::cmd_line_error_tp &e) {
        assert(e.what() == string("Usage: some_app_name input_file.in"));
    }

    // Too many parameters:
    try {
        int argc = 3;
        char *argv[] = {"some_app_name", "hihi", "huhu"};
        trasim_tp trasim(argc, argv);
        assert(false);
    } catch (const trasim_tp::cmd_line_error_tp &e) {
        assert(e.what() == string("Usage: some_app_name input_file.in"));
    }

    // Non-existent input file:
    try {
        int argc = 2;
        char *argv[] = {"some_app_name", "non-existent.ini"};
        trasim_tp trasim(argc, argv);
        assert(false);
    } catch (const params_tp::open_file_error_tp &e) {
        assert(e.what() == string("Cannot open file \"non-existent.ini\"."));
    }

    // Unknown procedure:
    try {
        int argc = 2;
        char *argv[] = {"some_app_name", "test1-1.in"};
        trasim_tp trasim(argc, argv);
        assert(false);
    } catch (const trasim_tp::procedure_error_tp &e) {
        assert(e.what() == string("Unknown procedure \"blabla\"."));
    }

    // rcrd-file-name missing:
    try {
        int argc = 2;
        char *argv[] = {"some_app_name", "test1-2.in"};
        trasim_tp trasim(argc, argv);
        assert(false);
    } catch (const params_tp::get_key_error_tp &e) {
        assert(true);
    }
}

/** Tests running simulations and dumping of data. */
void test2() {
    int argc = 2;
    char *argv[] = {"some_app_name", "test2.in"};
    trasim_tp trasim(argc, argv);
    trasim.run();
}

/** Tests initialization of the random number generator and the storage of the
    seed in the misc file. */
void test3() {
    // random_seed contains junk text:
    char *argv1[] = {"some_app_name", "test3-1.in"};
    try {
        trasim_tp trasim1(2, argv1);
        assert(false);
    } catch (const trasim_tp::invalid_seed_error_tp &e) {
        assert(e.what() == string("The random seed \"aslkdj\" is invalid."));
    }

    // random_seed is a floating point number:
    char *argv2[] = {"some_app_name", "test3-2.in"};
    try {
        trasim_tp trasim2(2, argv2);
        assert(false);
    } catch (const trasim_tp::invalid_seed_error_tp &e) {
        assert(e.what() == string("The random seed \"12.34\" is invalid."));
    }

    // random_seed is a negative number:
    char *argv3[] = {"some_app_name", "test3-3.in"};
    try {
        trasim_tp trasim3(2, argv3);
        assert(false);
    } catch (const trasim_tp::invalid_seed_error_tp &e) {
        assert(e.what() == string("The random seed \"-54\" is invalid."));
    }

    // random_seed is empty:
    char *argv4[] = {"some_app_name", "test3-4.in"};
    trasim_tp trasim4(2, argv4);
    trasim4.run();
    params_tp misc_params4("test3-4.misc");
    assert(misc_params4.get<string>("trasim:random-seed") == "0");

    // random_seed is missing:
    char *argv5[] = {"some_app_name", "test3-5.in"};
    trasim_tp trasim5(2, argv5);
    trasim5.run();
    params_tp misc_params5("test3-5.misc");
    assert(misc_params5.get<string>("trasim:random-seed") == "0");

    // random_seed is a large number:
    char *argv6[] = {"some_app_name", "test3-6.in"};
    trasim_tp trasim6(2, argv6);
    trasim6.run();
    params_tp misc_params6("test3-6.misc");
    assert(misc_params6.get<string>("trasim:random-seed") == "2147500032");
    // Compares the first four random numbers:
    assert(mt19937.genrand_int32() == 3334606378ul &&
           mt19937.genrand_int32() == 1848588718ul &&
           mt19937.genrand_int32() == 1726629441ul &&
           mt19937.genrand_int32() == 112582719ul);

    // random_seed is "time":
    char *argv7[] = {"some_app_name", "test3-7.in"};
    trasim_tp *trasim7 = new trasim_tp(2, argv7);
    trasim7->run();
    params_tp *misc_params7 = new params_tp("test3-7.misc");
    unsigned long s7 = misc_params7->get<unsigned long>("trasim:random-seed");
    // The following loop should stop at some point, if seeding by time works:
    do {
        delete trasim7;
        trasim7 = new trasim_tp(2, argv7);
        trasim7->run();
        delete misc_params7;
        misc_params7 = new params_tp("test3-7.misc");
    } while (misc_params7->get<unsigned long>("trasim:random-seed") == s7);
    delete misc_params7;
    delete trasim7;
}

int main() {
    vector<test_tp> tests;
    tests.push_back(test1);
    tests.push_back(test2);
    tests.push_back(test3);
    return run_tests(tests, "trasim");
}
