#include <measurement.hpp>
#include <h2.hpp>
#include <params.hpp>
#include <uniform-tracking.hpp>
#include <uniform-tracking_guess.hpp>
#include <uniform-tracking_const-guess.hpp>
#include <uniform-tracking_aks-guess.hpp>
#include <../test-suite/suite.hpp>
#include <iostream>
#include <cmath>
#include <complex>
#include <exception>
#include <cstdlib>
#include <cassert>
#include <cstring>
using namespace std;

typedef complex<double> dx;

/** Tests some error messages. */
static void test1() {
    ostringstream dummy_out; // We don't observe this data

    // Unknown guess:
    params_tp params1("test1-1.in");
    try {
        uniform_tracking_tp ut(params1, dummy_out, dummy_out);
        assert(false);
    } catch (const uniform_tracking_tp::wrong_guess_error_tp &e) {
        assert(true);
    }

    // tau <= 0
    params_tp params2("test1-2.in");
    try {
        uniform_tracking_tp ut(params2, dummy_out, dummy_out);
        assert(false);
    } catch (const uniform_tracking_tp::tau_le_0_error_tp &e) {
        assert(true);
    }

    // n_rec_points = 0 < 1
    params_tp params3("test1-3.in");
    try {
        uniform_tracking_tp ut(params3, dummy_out, dummy_out);
        assert(false);
    } catch (const uniform_tracking_tp::n_rec_points_too_small_error_tp &e) {
        assert(true);
    }

    // n_rec_points = -5 < 1
    params_tp params4("test1-4.in");
    try {
        uniform_tracking_tp ut(params4, dummy_out, dummy_out);
        assert(false);
    } catch (const uniform_tracking_tp::n_rec_points_too_small_error_tp &e) {
        assert(true);
    }

    // start_delay = -2 < 0
    params_tp params5("test1-5.in");
    try {
        uniform_tracking_tp ut(params5, dummy_out, dummy_out);
        assert(false);
    } catch (const uniform_tracking_tp::start_delay_too_small_error_tp &e) {
        assert(true);
    }
}

/** Tests output of output of miscellaneous data (especially eigenvalues of
    measurement operators). */
static void test2() {
    ostringstream misc_out;
    misc_out.precision(4);
    ostringstream rcrd_out;
    string misc_out_exp; // Expected output of uniform tracking (output of
                         // guesses might be omitted here)
    string rcrd_out_exp; // Expected output of (header of) measurement data

    // One measurement operator
    params_tp params1("test2-1.in");
    uniform_tracking_tp ut1(params1, misc_out, rcrd_out);
    misc_out_exp = 
        "\n\n"
        "[uniform-tracking]\n"
        "\n"
        "# Eigenvalues of effect operators and related data:\n"
        "lambda0_0 = (1,0)\n"
        "lambda0_1 = (1,0)\n"
        "Delta_lambda0_abs = 0\n"
        "mean_lambda0_abs = 1\n";
    rcrd_out_exp = "# t/(dim(tracking:E_0)/h_bar) |c_1_tilde|^2 |c_1|^2 g_1\n";
    assert(strncmp(misc_out.str().c_str(), misc_out_exp.c_str(), 
                   strlen(misc_out_exp.c_str())) == 0); // Note that we only
                                                        // compare the relevant
                                                        // first part
    assert(rcrd_out.str() == rcrd_out_exp);

    // Two measurement operators
    params_tp params2("test2-2.in");
    misc_out.str("");
    uniform_tracking_tp ut2(params2, misc_out, rcrd_out);
    misc_out_exp = 
        "\n\n"
        "[uniform-tracking]\n"
        "\n"
        "# Eigenvalues of effect operators and related data:\n"
        "lambda0_0 = (0.46,0)\n"
        "lambda0_1 = (0.54,0)\n"
        "Delta_lambda0_abs = 0.08\n"
        "mean_lambda0_abs = 0.5\n"
        "lambda1_0 = (0.46,0)\n"
        "lambda1_1 = (0.54,0)\n"
        "Delta_lambda1_abs = 0.08\n"
        "mean_lambda1_abs = 0.5\n";
    assert(strncmp(misc_out.str().c_str(), misc_out_exp.c_str(), 
                   strlen(misc_out_exp.c_str())) == 0);

    // Two measurement operators
    params_tp params3("test2-3.in");
    misc_out.str("");
    uniform_tracking_tp ut3(params3, misc_out, rcrd_out);
    misc_out_exp = 
        "\n\n"
        "[uniform-tracking]\n"
        "\n"
        "# Eigenvalues of effect operators and related data:\n"
        "lambda0_0 = (0.46,0)\n"
        "lambda0_1 = (0.54,0)\n"
        "Delta_lambda0_abs = 0.08\n"
        "mean_lambda0_abs = 0.5\n"
        "lambda1_0 = (0.46,0)\n"
        "lambda1_1 = (0.54,0)\n"
        "Delta_lambda1_abs = 0.08\n"
        "mean_lambda1_abs = 0.5\n";
    assert(strncmp(misc_out.str().c_str(), misc_out_exp.c_str(), 
                   strlen(misc_out_exp.c_str())) == 0);

    // Same as previous example but with swapped measurement operators (note
    // that this doesn't affect the order of the eigenvalues and consequently
    // doesn't affect the output)
    params_tp params4("test2-4.in");
    misc_out.str("");
    uniform_tracking_tp ut4(params4, misc_out, rcrd_out);
    misc_out_exp = 
        "\n\n"
        "[uniform-tracking]\n"
        "\n"
        "# Eigenvalues of effect operators and related data:\n"
        "lambda0_0 = (0.46,0)\n"
        "lambda0_1 = (0.54,0)\n"
        "Delta_lambda0_abs = 0.08\n"
        "mean_lambda0_abs = 0.5\n"
        "lambda1_0 = (0.46,0)\n"
        "lambda1_1 = (0.54,0)\n"
        "Delta_lambda1_abs = 0.08\n"
        "mean_lambda1_abs = 0.5\n";
    assert(strncmp(misc_out.str().c_str(), misc_out_exp.c_str(), 
                   strlen(misc_out_exp.c_str())) == 0);

    // Three measurement operators
    params_tp params5("test2-5.in");
    misc_out.str("");
    uniform_tracking_tp ut5(params5, misc_out, rcrd_out);
    misc_out_exp = 
        "\n\n"
        "[uniform-tracking]\n"
        "\n"
        "# Eigenvalues of effect operators and related data:\n"
        "lambda0_0 = (0.1,0)\n"
        "lambda0_1 = (0.3,0)\n"
        "Delta_lambda0_abs = 0.2\n"
        "mean_lambda0_abs = 0.2\n"
        "lambda1_0 = (0.5,0)\n"
        "lambda1_1 = (0.6,0)\n"
        "Delta_lambda1_abs = 0.1\n"
        "mean_lambda1_abs = 0.55\n"
        "lambda2_0 = (0.2,0)\n"
        "lambda2_1 = (0.3,0)\n"
        "Delta_lambda2_abs = 0.1\n"
        "mean_lambda2_abs = 0.25\n";
    assert(strncmp(misc_out.str().c_str(), misc_out_exp.c_str(), 
                   strlen(misc_out_exp.c_str())) == 0);
}

/** Tests the \t predefined-outcomes input file parameter and of the
    function \t outcome(n) which returns outcome number \p n. */
static void test3() {
    ostringstream misc_out, rcrd_out;

    // "predefined-outcomes" empty means no measurements should be read:
    params_tp params1("test3-1.in");
    uniform_tracking_tp ut1(params1, misc_out, rcrd_out);
    ut1.track(); // Should run without trouble

    // "predefined-outcomes" contain a non existent file name:
    params_tp params2("test3-2.in");
    try {
        uniform_tracking_tp ut2(params2, misc_out, rcrd_out);
        assert(false);
    } catch (const uniform_tracking_tp::
             cannot_read_predef_outcomes_error_tp &e) 
    {
        assert(e.what() == string("Reading outcomes from "
                                  "\"test3-2.outc\" failed."));
    }

    // The file specified by "predefined-outcomes" contains too few values:
    params_tp params3("test3-3.in");
    try {
        uniform_tracking_tp ut3(params3, misc_out, rcrd_out);
        assert(false);
    } catch (const uniform_tracking_tp::
             cannot_read_predef_outcomes_error_tp &e) 
    {
        assert(e.what() == string("Reading outcomes from "
                                  "\"test3-3.outc\" failed."));
    }

    // The file specified by "predefined-outcomes" contains non integer values:
    params_tp params4("test3-4.in");
    try {
        uniform_tracking_tp ut4(params4, misc_out, rcrd_out);
        assert(false);
    } catch (const uniform_tracking_tp::
             cannot_read_predef_outcomes_error_tp &e) 
    {
        assert(e.what() == string("Reading outcomes from "
                                  "\"test3-4.outc\" failed."));
    }

    // The file specified by "predefined-outcomes" contains values that are out
    // of range (1 measurement operator):
    params_tp params5("test3-5.in");
    try {
        uniform_tracking_tp ut5(params5, misc_out, rcrd_out);
        assert(false);
    } catch (const uniform_tracking_tp::
             outcome_out_of_range_error_tp &e) 
    {
        assert(e.what() == 
               string("Outcome 2 read from \"test3-5.outc\" is out of range."));
    }

    // The file specified by "predefined-outcomes" contains values that are out
    // of range (2 measurement operators):
    params_tp params6("test3-6.in");
    try {
        uniform_tracking_tp ut6(params6, misc_out, rcrd_out);
        assert(false);
    } catch (const uniform_tracking_tp::outcome_out_of_range_error_tp &e) {
        assert(e.what() == 
               string("Outcome 4 read from \"test3-6.outc\" is out of range."));
    }

    // The file specified by "predefined-outcomes" is good and contains more
    // entries than necessary (1 meas. operator):
    params_tp params7("test3-7.in");
    uniform_tracking_tp ut7(params7, misc_out, rcrd_out);
    ut7.track();    
    assert(ut7.outcome(0) == 0 && ut7.outcome(1) == 0 && ut7.outcome(2) == 0 &&
           ut7.outcome(3) == 0 && ut7.outcome(4) == 0 && ut7.outcome(5) == 0);

    // The file specified by "predefined-outcomes" is good (2 meas. operators):
    params_tp params8("test3-8.in");
    uniform_tracking_tp ut8(params8, misc_out, rcrd_out);
    ut8.track();    
    assert(ut8.outcome(0) == 0 && ut8.outcome(1) == 1 && ut8.outcome(2) == 1 &&
           ut8.outcome(3) == 0 && ut8.outcome(4) == 1 && ut8.outcome(5) == 1 &&
           ut8.outcome(6) == 0 && ut8.outcome(7) == 0 && ut8.outcome(8) == 0 &&
           ut8.outcome(9) == 0 && ut8.outcome(10) == 1 && ut8.outcome(11) == 1);
}

int main() {
    vector<test_tp> tests;
    tests.push_back(test1);
    tests.push_back(test2);
    tests.push_back(test3);
    // For more tests see also the tests for uniform-tracking_const-guess,
    // uniform-tracking_aks-guess, etc..
    return run_tests(tests, "uniform-tracking");
}
