#include <measurement.hpp>
#include <h2.hpp>
#include <params.hpp>
#include <uniform-tracking.hpp>
#include <uniform-tracking_guess.hpp>
#include <../test-suite/suite.hpp>
#include <iostream>
#include <cmath>
#include <complex>
#include <exception>
#include <cstdlib>
#include <cassert>
#include <cstring>
using namespace std;

typedef complex<double> dx;

/** Simple tracking procedure. Constant guess. If \p no_time_evol == true then
    we assume that no time evolution takes place. \p file_name is the name of
    the file from which the parameters should be read. Note that this file
    needs to contain a description for a tracking procedure with measurement
    operators "that don't measure" (the unit matrix is such an operator) and
    with constant guess. */
static void test_constant(const string &file_name, 
                          const operator_tp &U_expected,
                          int start_delay_expected,
                          double g_expected,
                          const vector<double> &c_1_absq_expected,
                          const string &rcrd_expected,
                          bool no_time_evol = false) 
{
    params_tp params(file_name);

    ostringstream misc_stream;
    ostringstream rcrd_stream;
    rcrd_stream.precision(4);
    uniform_tracking_tp ut(params, misc_stream, rcrd_stream);
    assert(approx_eq(ut.U(), U_expected, aprx_a));
    vector<operator_tp> M;
    M.push_back(operator_tp::one);
    measurement_tp measurement_expected(M);
    assert(approx_eq(ut.measurement(), measurement_expected, aprx_a));

    ut.track();

    if (no_time_evol)
        // psi doesn't change because of no measurements and no time evolution:
        assert(approx_eq(ut.psi(), ut.psi_init(), aprx_a, 1e-2));
        
    // Check g_1, c_1, and c_1_tilde (Note that due to numerical
    // inaccuracies the norm of psi and psi_tilde isn't always exactly one
    // and therefore we can't do comparisons with ultra low tolerance. This
    // error, however, doesn't build up since psi is normalized after each
    // measurement and psi_tilde is normalized after unitary evolution):
    vector<tracking_tp::record_tp>::const_iterator i = ut.recording().begin();
    vector<double>::const_iterator j = c_1_absq_expected.begin();
    for (int l = 0; i != ut.recording().end(); ++i, ++j, ++l) {
        assert(approx_eq(i->g_1_mb, (l < start_delay_expected) ? 
                         0 : g_expected));
        assert(approx_eq(i->c_1_absq_mb, *j, aprx_r, 1e-4));
        assert(approx_eq(i->c_1_tilde_absq_mb, *j, aprx_r, 1e-4));
    }
    
    // Test dumping of data after tracking (As in the test above, due to
    // numerical inaccuracies we are not comparing values at ultra high
    // precision.):
    assert(rcrd_stream.str() == rcrd_expected);
    rcrd_stream.str("");
}

static void test1() {
    vector<double> c_1_absq_expected1;
    string rcrd_expected1;
    rcrd_expected1 = "# t/(dim(tracking:E_0)/h_bar) |c_1_tilde|^2 |c_1|^2 g_1\n";
    for (int i = 0; i < 4; ++i) {
        c_1_absq_expected1.push_back(0.0835712);
        rcrd_expected1 += stringify(i*0.1)+"\t0.08357\t0.08357\t0.73\n";
    }
    test_constant("test1-1.in", operator_tp::one, 0, 0.73, c_1_absq_expected1, 
                  rcrd_expected1, true);

    operator_tp U2(dx(-0.247265,0.72377), dx(0.609623,0.208269),
                   dx(0.609623,0.208269), dx(-0.247265,0.72377));
    vector<double> c_1_absq_expected2;
    c_1_absq_expected2.push_back(0.0835712);
    c_1_absq_expected2.push_back(0.270836);
    c_1_absq_expected2.push_back(0.838577);
    string rcrd_expected2 =
        "# t/(dim(tracking:E_0)/h_bar) |c_1_tilde|^2 |c_1|^2 g_1\n"
        "0\t0.08357\t0.08357\t0.24\n"
        "0.1\t0.2708\t0.2708\t0.24\n"
        "0.2\t0.8385\t0.8385\t0.24\n";
    test_constant("test1-2.in", U2, 0, 0.24, c_1_absq_expected2, 
                  rcrd_expected2);

    // Almost the same test as above but with different guess, more log points,
    // and start delay:
    operator_tp U3(dx(0.875711,0.450354), dx(0.0796264,-0.154833),
                   dx(0.0796264,-0.154833), dx(0.875711,0.450354));
#include "test1-3_expected.cpp"
    test_constant("test1-3.in", U3, 5, 0.76, c_1_absq_expected3, 
                  rcrd_expected3);
}

int main() {
    vector<test_tp> tests;
    tests.push_back(test1);
    return run_tests(tests, "uniform-tracking_const-guess");
}
