#include <measurement.hpp>
#include <h2.hpp>
#include <params.hpp>
#include <uniform-tracking.hpp>
#include <uniform-tracking_guess.hpp>
#include <uniform-tracking_min-dev-xrot-guess.hpp>
#include <../test-suite/suite.hpp>
#include <iostream>
#include <cmath>
#include <complex>
#include <exception>
#include <cstdlib>
#include <cassert>
#include <cstring>
#include <vector>
#include <gmpxx.h>
using namespace std;

typedef complex<double> dx;
dx ii(0,1);
typedef uniform_tracking_tp::min_dev_xrot_guess_tp::a_tp a_tp;

/** Converts complex_mpf_tp to complex<double>. */
static inline complex<double> 
to_complex_double(const complex_mpf_tp &x) {
    return complex<double>(real(x).get_d(), imag(x).get_d());
}

/** Just tests reading of parameters file. */
static void test1() {
    ostringstream some_stream;

    // Should not produce any errors and use the MPF default precision:
    string stream_misc_expected1 = 
        "\n\n[minimum-deviation-xrot-guess]\n"
        "precision-used = "+stringify(mpf_get_default_prec())+"\n";
    params_tp params1("test1-1.in");
    uniform_tracking_tp ut1(params1, some_stream, some_stream);
    ostringstream out1;
    out1.precision(4);
    ut1.guess().stream_misc(out1);
    assert(strncmp(out1.str().c_str(), stream_misc_expected1.c_str(), 
                   stream_misc_expected1.size()) == 0);
}

/** Tests the creation of the vector {I_sc2(0,N0-0), I_sc(1,N0-1),
    I_sc2(2,N0-2), ..., I_sncm(N0,N0-N0)} (see 0306181630). */
static void test2() {
	// Dummies:
    params_tp params("test2.in");
    vector<operator_tp> M;
    M.push_back(operator_tp::one);
    measurement_tp meas(M); // Measurement

	// N0 = 0, ..., 4:
    int max_N = 5; // Number of measurements
    uniform_tracking_tp::min_dev_xrot_guess_tp g1(params, meas, max_N);
    double I_sc2_expected[5][5] = 
        {{6.283}, 
         {3.142, 3.142}, 
         {2.356, 0.7854, 2.356}, 
         {1.963, 0.3927, 0.3927, 1.963}, 
         {1.718, 0.2454, 0.1473, 0.2454, 1.718}};
    for (int N0 = 0; N0 <= max_N-1; ++N0) {
        g1.new_outcome(0);
        g1.guess();
        const mpf_vect_tp &I_sc2 = g1.I_sc2();
        for (int q = 0; q <= N0; ++q)
            assert(approx_eq(I_sc2[q].get_d(), I_sc2_expected[N0][q], aprx_r, 
                             1e-3));
    }

    max_N = 100;
    uniform_tracking_tp::min_dev_xrot_guess_tp g2(params, meas, max_N);
    double I_sc2_expected_0_99 = 0.35582707074661035;
	double I_sc2_expected_3_99 = 7.198998746575482e-7;
	double I_sc2_expected_49_99 = 7.969470115145116e-31;
    for (int N0 = 0; N0 <= max_N-1; ++N0)
        g2.new_outcome(0), g2.guess();
	assert(approx_eq(g2.I_sc2()[0].get_d(), I_sc2_expected_0_99, aprx_r, 1e-3));
	assert(approx_eq(g2.I_sc2()[3].get_d(), I_sc2_expected_3_99, aprx_r, 1e-3));
	assert(approx_eq(g2.I_sc2()[49].get_d(), I_sc2_expected_49_99, aprx_r, 1e-3));
}

/** Executes the guess \p max_N times and compares the generated values of the
    generated a_jkn with a_expected. \p m are the measurement results that
    should be fed into the guess. \p meas is the measurement used for the
    guess.

    \p a_expected = 
    {a^(1)_000, a^(1)_010, a^(1)_100, a^(1)_110,
     a^(2)_000, a^(2)_010, a^(2)_000, a^(2)_110, 
     a^(2)_001, a^(2)_011, a^(2)_101, a^(2)_111,
     ...,
     a^(max_N)_000, a^(max_N)_010, a^(max_N)_000, a^(max_N)_110, 
     a^(max_N)_001, a^(max_N)_011, a^(max_N)_001, a^(max_N)_111, ...,
     a^(max_N)_00max_N0, a^(max_N)_01max_N0, 
     a^(max_N)_00max_N0, a^(max_N)_11max_N0} */
static void test3_verify_a(int max_N, measurement_tp &meas,
                           const int *m, const dx *a_expected) 
{
    params_tp params("test3.in"); // Needed only for initialization of the
                                  // guess, the contents of test3.in are
                                  // ignored.
    uniform_tracking_tp::min_dev_xrot_guess_tp g(params, meas, max_N);
    int l = 0; // Position in a_expected
    for (int N0 = 0; N0 <= max_N-1; ++N0) {
        g.new_outcome(m[N0]);
        g.guess();
        const uniform_tracking_tp::min_dev_xrot_guess_tp::a_tp &a = g.a();
        for (int n = 0; n <= N0; ++n)
            for (int j = 0; j <= 1; ++j)
                for (int k = 0; k <= 1; ++k)
                    assert(approx_eq(to_complex_double(a(j,k,n)),
                                     a_expected[l++],
                                     aprx_r, 1e-3));
    }
}

/** Tests the creation of the a_jkn (see 0307091130). */
// QQQ: Add test for the guess itself
static void test3() {
    int max_N; // Number of measurements
    vector<operator_tp> Ms; // Measurement operators

    // No measurements:
    max_N = 4;
    Ms.push_back(operator_tp::one);
    measurement_tp meas1(Ms);
    int m1[] = {0,0,0,0};
    dx a_expected1[] = { 
#include "test3_a_expected1"
    };
    test3_verify_a(max_N, meas1, m1, a_expected1);

    // Minimal measurements with the operators sqrt(0.3)|0><0|+sqrt(0.7)|1><1|,
    // sqrt(0.7)|0><0|+sqrt(0.3)|1><1|:
    max_N = 4;
    Ms.clear();
    Ms.push_back(operator_tp(0.5477225575051661,0.,0.,0.8366600265340756));
    Ms.push_back(operator_tp(0.8366600265340756,0.,0.,0.5477225575051661));
    measurement_tp meas2(Ms);
    int m2[] = {0,1,0,0};
    dx a_expected2[] = {
#include "test3_a_expected2"
    };
    test3_verify_a(max_N, meas2, m2, a_expected2);

    // Nonminimal measurements:
    max_N = 4;
    Ms.clear();
    Ms.push_back(operator_tp(dx(0.5106277157878547,0.41691166777458855),
                             dx(0.1616447389998824,0.11425562152679138),
                             dx(-0.19813952774662302,-0.16181655619802346),
                             dx(0.4169784667908918,0.29486559033422266)));
    Ms.push_back(operator_tp(dx(0.5895643382415912,0.00018249830643637743),
                             dx(0.11411813355002595,0.22917509372992317),
                             dx(-0.2289403870724554,0.0004694133149356004),
                             dx(0.2949190428506228,0.5894730890883729)));
    Ms.push_back(operator_tp(dx(0.20841020931068516,0.20853925510072138),
                             dx(-0.00016596266908511118,0.16188530018640618),
                             dx(-0.08102563142774565,-0.08069370608957543),
                             dx(0.00006452289501811358,0.41688494151638844)));
    measurement_tp meas3(Ms);
    int m3[] = {1,2,0,1};
    dx a_expected3[] = {
#include "test3_a_expected3"
    };
    test3_verify_a(max_N, meas3, m3, a_expected3);
}

/** Executes the guess \p max_N times and compares the generated values and \p
    pi2_E_tilde_bar_00, \p pi2_E_tilde_bar_11, and \p guess with \p
    pi2_E_tilde_bar_00_expected, \p pi2_E_tilde_bar_11_expected and \p
    guess_expected. \p m are the measurement results that are fed into the
    guess. \p meas is the measurement used for the guess. \p params_file_name
    is the name of the parameters file used for the guess. */
static void test4_verify(int max_N, measurement_tp &meas,
                         const int *m,
                         double *pi2_E_tilde_bar_00_expected, 
                         double *pi2_E_tilde_bar_11_expected,
                         double *guess_expected,
                         const string &params_file_name)
{
    params_tp params(params_file_name);
    uniform_tracking_tp::min_dev_xrot_guess_tp g(params, meas, max_N);
    for (int j = 0; j < max_N; ++j) {
        g.new_outcome(m[j]);
        double guess = g.guess();
        assert(approx_eq(g.pi2_E_tilde_bar_00().get_d(),
                         pi2_E_tilde_bar_00_expected[j], aprx_r, 1e-3));
        assert(approx_eq(g.pi2_E_tilde_bar_11().get_d(),
                         pi2_E_tilde_bar_11_expected[j], aprx_r, 1e-3));
        assert(approx_eq(guess, guess_expected[j], aprx_a, 1e-3));
    }
}

/** Tests the creation of 2pi_E_tilde_bar_00 and 2pi_E_tilde_bar_11 (see
    0307091720). */
static void test4() {
    int max_N; // Number of measurements
    vector<operator_tp> Ms; // Measurement operators

    // No measurements:
    max_N = 4;
    Ms.push_back(operator_tp::one);
    measurement_tp meas1(Ms);
    int m1[] = {0,0,0,0};
    double pi2_E_tilde_bar_00_expected1[] = 
        {6.28319, 6.28319, 6.28319, 6.28319};
    double pi2_E_tilde_bar_11_expected1[] = 
        {6.28319, 6.28319, 6.28319, 6.28319};
    double guess_expected1[] = {0.5, 0.5, 0.5, 0.5};
    test4_verify(max_N, meas1, m1, 
                 pi2_E_tilde_bar_00_expected1, pi2_E_tilde_bar_11_expected1,
                 guess_expected1, "test4-1.in");

    // Minimal measurements with the operators sqrt(0.3)|0><0|+sqrt(0.7)|1><1|,
    // sqrt(0.7)|0><0|+sqrt(0.3)|1><1|:
    max_N = 4;
    Ms.clear();
    Ms.push_back(operator_tp(0.5477225575051661,0.,0.,0.8366600265340756));
    Ms.push_back(operator_tp(0.8366600265340756,0.,0.,0.5477225575051661));
    measurement_tp meas2(Ms);
    int m2[] = {0,1,0,0};
    double pi2_E_tilde_bar_00_expected2[] = 
        {1.88496, 2.19911, 0.463371, 0.248667};
    double pi2_E_tilde_bar_11_expected2[] = 
        {4.39823, 0.942478, 1.11792, 0.526678};
    double guess_expected2[] = {0.7, 0.3, 0.706966, 0.679282};
    test4_verify(max_N, meas2, m2, 
                 pi2_E_tilde_bar_00_expected2, pi2_E_tilde_bar_11_expected2,
                 guess_expected2, "test4-1.in");

    // Nonminimal measurements:
    max_N = 4;
    Ms.clear();
    Ms.push_back(operator_tp(dx(0.5106277157878547,0.41691166777458855),
                             dx(0.1616447389998824,0.11425562152679138),
                             dx(-0.19813952774662302,-0.16181655619802346),
                             dx(0.4169784667908918,0.29486559033422266)));
    Ms.push_back(operator_tp(dx(0.5895643382415912,0.00018249830643637743),
                             dx(0.11411813355002595,0.22917509372992317),
                             dx(-0.2289403870724554,0.0004694133149356004),
                             dx(0.2949190428506228,0.5894730890883729)));
    Ms.push_back(operator_tp(dx(0.20841020931068516,0.20853925510072138),
                             dx(-0.00016596266908511118,0.16188530018640618),
                             dx(-0.08102563142774565,-0.08069370608957543),
                             dx(0.00006452289501811358,0.41688494151638844)));
    measurement_tp meas3(Ms);
    int m3[] = {1,2,0,1};
    double pi2_E_tilde_bar_00_expected3[] = 
        {2.59577, 0.334181, 0.229522, 0.0722066};
    double pi2_E_tilde_bar_11_expected3[] = 
        {3.05909, 0.514082, 0.109499, 0.0795544};
    double guess_expected3[] = {0.540966, 0.606041, 0.322986, 0.524209};
    test4_verify(max_N, meas3, m3, 
                 pi2_E_tilde_bar_00_expected3, pi2_E_tilde_bar_11_expected3,
                 guess_expected3, "test4-1.in");
}

/** Tests of precision and divergency. */
static void test5() {
    ostringstream some_stream;
    int measurements_performed;

    // If the value for the precision is negative, then the default precision is
    // used instead:
    params_tp params1("test5-1.in");
    uniform_tracking_tp ut1(params1, some_stream, some_stream);
    const uniform_tracking_tp::min_dev_xrot_guess_tp *guess1 = dynamic_cast<
        const uniform_tracking_tp::min_dev_xrot_guess_tp*>(&ut1.guess());
    assert(guess1->precision() == mpf_get_default_prec());

    // If the value for the precision is zero, then the default precision is
    // used instead:
    params_tp params2("test5-2.in");
    uniform_tracking_tp ut2(params2, some_stream, some_stream);
    const uniform_tracking_tp::min_dev_xrot_guess_tp *guess2 = dynamic_cast<
        const uniform_tracking_tp::min_dev_xrot_guess_tp*>(&ut2.guess());
    assert(guess2->precision() == mpf_get_default_prec());

    // If the simulation stops after N measurements due to convergence with
    // default precision, then it is expected to run for at least N+1
    // measurements with twice the default precision. Also, if the default
    // precision is greater than 64, then N should be greater than 80, and
    // N2 should be greater than 140.
#include "test5_3_data"
    string test5_3a_data = test5_3_data_begin + 
        stringify(mpf_get_default_prec()) + test5_3_data_end;
    string_to_file("test5-3a.in", test5_3a_data);
    params_tp params3a("test5-3a.in");
    uniform_tracking_tp ut3a(params3a, some_stream, some_stream);
    unsigned long N = ut3a.track(); // N is the # of measurements
    string test5_3b_data = test5_3_data_begin + 
        stringify(2*mpf_get_default_prec()) + test5_3_data_end;
    string_to_file("test5-3b.in", test5_3b_data);
    params_tp params3b("test5-3b.in");
    uniform_tracking_tp ut3b(params3b, some_stream, some_stream);
    unsigned long N2 = ut3b.track();
    assert(N2 > N);
    if (mpf_get_default_prec() >= 64) {
        // On systems/implementations with lower default precision this test
        // would fail (it is actually doubtful that such systems/
        // implementations exist)
        assert(N > 80);
        assert(N2 > 140);
    }
}

/** Test that tracking with projective operators doesn't lead to floating point
    exceptions. */
static void test6() {
    ostringstream some_stream;
    params_tp params("test6.in");
    uniform_tracking_tp ut(params, some_stream, some_stream);
    int measurements_performed = ut.track(); // Should not cause an exception
    assert(measurements_performed == 10);
}

/** Tests of the minimum deviation guess for Rabi oszillations. */
int main() {
    vector<test_tp> tests;
    tests.push_back(test1);
    tests.push_back(test2);
    tests.push_back(test3);
    tests.push_back(test4);
    tests.push_back(test5);
    tests.push_back(test6);
    return run_tests(tests, "uniform-tracking_min-dev-xrot-guess");
}
